"""Tasks for chaining workflows in repod."""

from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from collections import defaultdict
from itertools import groupby
from logging import debug, info
from operator import attrgetter
from pathlib import Path
from re import sub
from shutil import copy2

from orjson import JSONEncodeError, dumps
from pydantic import AnyUrl, BaseModel, ValidationError, field_validator

from repod.action.check import (
    Check,
    DebugPackagesCheck,
    MatchingArchitectureCheck,
    MatchingFilenameCheck,
    PackagesNewOrUpdatedCheck,
    PacmanKeyPackagesSignatureVerificationCheck,
    PkgbasesVersionUpdateCheck,
    ReproducibleBuildEnvironmentCheck,
    SourceUrlCheck,
    StabilityLayerCheck,
    UniqueInRepoGroupCheck,
)
from repod.archive.archive import CopySourceDestination
from repod.common.enums import (
    ActionStateEnum,
    ArchitectureEnum,
    CompressionTypeEnum,
    FilesVersionEnum,
    PackageDescVersionEnum,
    PkgVerificationTypeEnum,
    RepoDirTypeEnum,
    RepoFileEnum,
    RepoTypeEnum,
)
from repod.common.models import FileName
from repod.config import PackageRepo, SystemSettings, UserSettings
from repod.config.defaults import ORJSON_OPTION
from repod.config.settings import UrlValidationSettings
from repod.errors import (
    RepoManagementFileError,
    RepoManagementFileNotFoundError,
    TaskError,
)
from repod.files import Package
from repod.files.buildinfo import Installed
from repod.repo import OutputPackageBase, SyncDatabase
from repod.repo.package import RepoDbTypeEnum, RepoFile
from repod.repo.package.repofile import relative_to_shared_base


def read_build_requirements_from_archive_dir(
    pkgbases: list[OutputPackageBase],
    archive_dir: Path | None,
    pkgs_in_archive: set[str],
) -> None:
    """Read build requirements of a list of OutputPackageBases from an archive directory.

    Parameters
    ----------
    pkgbases: list[OutputPackageBase]
        A list of OutputPackageBase instances for which to read build requirements
    archive_dir: Path | None
        An optional archive directory, from which to read package information
    pkgs_in_archive: set[str]
        A set of strings to which matching build requirements in the archive are appended
    """
    if not archive_dir:
        return

    for pkgbase in pkgbases:
        for pkgname, pkgver, architecture in Installed.as_models(
            pkgbase.buildinfo.installed  # type: ignore[attr-defined]
        ):
            requirement = f"{pkgname.pkgname}-{pkgver.pkgver}-{architecture.value}"
            if requirement not in pkgs_in_archive:
                pkg_archive_dir = archive_dir / pkgname.pkgname[0] / pkgname.pkgname
                paths = [path for path in pkg_archive_dir.glob(f"{requirement}*") if path.suffix != ".sig"]
                if len(paths) == 1:
                    try:
                        FileName(filename=paths[0].name)
                    except ValidationError as e:
                        raise TaskError(e)

                    pkgs_in_archive.add(requirement)


def read_build_requirements_from_management_repo_dirs(
    pkgbases: list[OutputPackageBase],
    management_directories: list[Path],
    pkgs_in_repo: set[str],
) -> None:
    """Read build requirements of a list of OutputPackageBases from management repository directories.

    Parameters
    ----------
    pkgbases: list[OutputPackageBase]
        A list of OutputPackageBase instances for which to read build requirements
    management_directories: list[Path]
        A list of management repository directories, from which to read package information
    pkgs_in_repo: set[str]
        A set of strings to which matching build requirements in the management repository directories are appended

    Raises
    ------
    TaskError
        If an error occurs while reading a file from the management repository directories
    """
    for pkgbase in pkgbases:
        for pkgname, pkgver, architecture in Installed.as_models(
            pkgbase.buildinfo.installed  # type: ignore[attr-defined]
        ):
            requirement = f"{pkgname.pkgname}-{pkgver.pkgver}-{architecture.value}"
            if requirement not in pkgs_in_repo:
                for directory in management_directories:
                    debug(
                        f"Searching for {pkgbase.base}'s build requirement "  # type: ignore[attr-defined]
                        f"{requirement} in management repository directory {directory}..."
                    )
                    file = directory / "pkgnames" / f"{pkgname.pkgname}.json"
                    if file.exists():
                        try:
                            current_pkgbase = asyncio.run(OutputPackageBase.from_file(path=file))
                        except RepoManagementFileError as e:
                            raise TaskError(e)

                        current_pkg_arch = [
                            ArchitectureEnum(pkg.arch)
                            for pkg in current_pkgbase.packages  # type: ignore[attr-defined]
                            if pkg.name == pkgname.pkgname
                        ][0]
                        debug(
                            f"Found {pkgname.pkgname}-"
                            f"{current_pkgbase.version}-{current_pkg_arch}..."  # type: ignore[attr-defined]
                        )

                        if (
                            current_pkgbase.version == pkgver.pkgver  # type: ignore[attr-defined]
                            and current_pkg_arch == architecture
                        ):
                            pkgs_in_repo.add(requirement)


def read_pkgbases_from_stability_layers(
    directory: Path,
    pkgbase_names: list[str],
    stability_layer_dirs: tuple[list[Path], list[Path]],
    current_pkgbases: list[OutputPackageBase],
    current_filenames: list[str],
    current_package_names: list[str],
    pkgbases_above: list[OutputPackageBase],
    pkgbases_below: list[OutputPackageBase],
) -> None:
    """Read the pkgbases from all available stability layers.

    Parameters
    ----------
    directory: Path
        The Path of the default directory in which to find pkgbase JSON files
    pkgbase_names: list[str]
        The names of pkgbases to read
    stability_layer_dirs: tuple[list[Path], list[Path]]
        A tuple of two lists of Paths, that represent the stability layers above (first list) and below (second list)
        the default layer.
    current_pkgbases: list[OutputPackageBase]
        A list of OutputPackageBase objects to which to append the objects read from the default directory
    current_filenames: list[str]
        A list of strings to which to append the filenames of current pkgbases to
    current_package_names: list[str]
        A list of strings to which to append the package names of current pkgbases to
    pkgbases_above: list[OutputPackageBase]
        A list of OutputPackageBase objects to which to append the objects read from the layers above the default
    pkgbases_below: list[OutputPackageBase]
        A list of OutputPackageBase objects to which to append the objects read from the layers below the default

    Raises
    ------
    RepoManagementFileError
        If OutputPackageBase.from_file raises
    """
    for name in pkgbase_names:
        current_pkgbase_file = directory / Path(name + ".json")
        if current_pkgbase_file.exists():
            current_pkgbase = asyncio.run(OutputPackageBase.from_file(current_pkgbase_file))
            current_filenames += [
                package.filename for package in current_pkgbase.packages  # type: ignore[attr-defined]
            ]
            current_package_names += [
                package.name for package in current_pkgbase.packages  # type: ignore[attr-defined]
            ]
            current_pkgbases.append(current_pkgbase)

        for path in stability_layer_dirs[0]:
            path_above = path / Path(name + ".json")
            if path_above.exists():
                pkgbases_above.append(asyncio.run(OutputPackageBase.from_file(path_above)))

        for path in stability_layer_dirs[1]:
            path_below = path / Path(name + ".json")
            if path_below.exists():
                pkgbases_below.append(asyncio.run(OutputPackageBase.from_file(path_below)))


class SourceDestination(BaseModel):
    """A model describing a source and a destination of a file.

    Attributes
    ----------
    source: Path
        The source file (Path must have '.tmp' suffix)
    destination: Path
        The destination file
    destination_backup: Path
        The Path of the backup of destination (Path must have '.bkp' suffix)
    backup_done: bool
        Whether the destination exists and whether a backup of it has been created (defaults to False)
    """

    source: Path
    destination: Path
    destination_backup: Path
    backup_done: bool = False

    @field_validator("source")
    @classmethod
    def validate_source(cls, path: Path) -> Path:
        """Validate the source attribute.

        Raises
        ------
        ValueError
            If source does not have a .tmp suffix,
            or if source is not an absolute Path

        Parameters
        ----------
        path: Path
            A Path to validate

        Returns
        -------
        Path
            A validated Path
        """
        if not path.is_absolute():
            raise ValueError(f"The path Path must be absolute, but {path} is not!")
        if not str(path).endswith(".tmp"):
            raise ValueError(f"The path Path must end with '.tmp', but {path} does not!")

        return path

    @field_validator("destination")
    @classmethod
    def validate_destination(cls, path: Path) -> Path:
        """Validate the destination attribute.

        Raises
        ------
        ValueError
            If destination has a .tmp or .bkp suffix,
            or if destination is not an absolute Path

        Parameters
        ----------
        path: Path
            A Path to validate

        Returns
        -------
        Path
            A validated Path
        """
        if not path.is_absolute():
            raise ValueError(f"The destination Path must be absolute, but {path} is not!")
        if str(path).endswith(".tmp"):
            raise ValueError(f"The destination Path must not end with '.tmp', but {path} does!")
        if str(path).endswith(".bkp"):
            raise ValueError(f"The destination Path must not end with '.bkp', but {path} does!")

        return path

    @field_validator("destination_backup")
    @classmethod
    def validate_destination_backup(cls, path: Path) -> Path:
        """Validate the destination_backup attribute.

        Raises
        ------
        ValueError
            If destination does not have .bkp suffix,
            or if destination_backup is not an absolute Path

        Parameters
        ----------
        path: Path
            A Path to validate

        Returns
        -------
        Path
            A validated Path
        """
        if not path.is_absolute():
            raise ValueError(f"The destination_backup Path must be absolute, but {path} is not!")
        if not str(path).endswith(".bkp"):
            raise ValueError(f"The destination_backup Path must end with '.bkp', but {path} does not!")

        return path


class Task(ABC):
    """An abstract base class to describe an operation.

    Tasks are Callables, that are used to run an operation (e.g. on an input) and may have pre and post checks. A Task
    tracks its own state, which indicates whether it ran successfully or not, using a member of ActionStateEnum.

    When deriving from Task, the do() and undo() methods must be implemented.

    The do() method is automatically run in __call__() and is expected to return either ActionStateEnum.SUCCESS_TASK or
    ActionStateEnum.FAILED_TASK, depending on whether the Task finished successfully or failed (respectively).

    The undo() method must undo all actions that have been done in do(). The method is expected to reset a Task's state
    property back to ActionStateEnum.NOT_STARTED or ActionStateEnum.FAILED_UNDO_TASK and return its state property,
    depending on whether the undo operation finished successfully or failed (respectively).

    Attributes
    ----------
    dependencies: list[Task] | None
        An optional list of Task instances that are run before this task and its pre_checks (defaults to None)
    pre_checks: list[Check]
        A list of Check instances that are called before the Task is called
    post_checks: list[Check]
        A list of Check instances that are called after the Task is run
    state: ActionStateEnum
        A member of ActionStateEnum indicating whether the Task is unstarted, started, failed, failed in any of the pre
        or post checks or successfully finished (defaults to ActionStateEnum.NOT_STARTED)
    """

    dependencies: list[Task] = []
    pre_checks: list[Check] = []
    post_checks: list[Check] = []
    state: ActionStateEnum = ActionStateEnum.NOT_STARTED

    def __call__(self) -> ActionStateEnum:  # pragma: no cover
        """Call a Task.

        A Task has the following call order:
        - all of its dependency Tasks
        - the Checks listed in pre_checks
        - its own do() method
        - the Checks listed in post_checks

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.FAILED_DEPENDENCY if any of the dependency Tasks fails,
            ActionStateEnum.SUCCESS if the Task executed successfully (or is run again after running successfully)
            ActionStateEnum.FAILED_PRE_CHECK if any of the Checks in pre_checks fails,
            ActionStateEnum.FAILED_TASK if the do() method of the Task fails,
            ActionStateEnum.FAILED_POST_CHECK if  any of the Checks in post_checks fails,
        """
        for dependency in self.dependencies:
            if dependency() != ActionStateEnum.SUCCESS:
                self.state = ActionStateEnum.FAILED_DEPENDENCY
                return self.state

        if self.state == ActionStateEnum.SUCCESS:
            return self.state

        self.state = ActionStateEnum.STARTED

        for check in self.pre_checks:
            if check() != ActionStateEnum.SUCCESS:
                self.state = ActionStateEnum.FAILED_PRE_CHECK
                return self.state

        if self.do() != ActionStateEnum.SUCCESS_TASK:
            return self.state

        for check in self.post_checks:
            if check() != ActionStateEnum.SUCCESS:
                self.state = ActionStateEnum.FAILED_POST_CHECK
                return self.state

        self.state = ActionStateEnum.SUCCESS
        return self.state

    @abstractmethod
    def do(self) -> ActionStateEnum:  # pragma: no cover
        """Run the Task's operation.

        This runs the Task's operation and sets its state property to ActionStateEnum.SUCCESS_TASK or
        ActionStateEnum.FAILED_TASK, depending on whether it runs successful or not.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise.
        """
        pass

    @abstractmethod
    def undo(self) -> ActionStateEnum:  # pragma: no cover
        """Undo the Task's operation.

        This runs an operation to undo any actions done in a Task's do() call, sets the Task's state property to
        ActionStateEnum.NOT_STARTED if successful, otherwise to ActionStateEnum.FAILED_UNDO_TASK and returns the state.

        Before returning, implementations of this method are expected to call self.dependency_undo() to also undo any
        dependency tasks in reverse order.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        pass

    def dependency_undo(self) -> ActionStateEnum:  # pragma: no cover
        """Undo all dependency Tasks in reverse order.

        Returns
        -------
        ActionStateEnum
            The ActionStateEnum member before calling the method,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed
        """
        for dependency in reversed(self.dependencies):
            if dependency.undo() != ActionStateEnum.NOT_STARTED:
                self.state = ActionStateEnum.FAILED_UNDO_DEPENDENCY

        return self.state

    def is_done(self) -> bool:  # pragma: no cover
        """Return the done state of the Task as a boolean value.

        Returns
        -------
        bool
            True if the Task is done (self.state is ActionStateEnum.SUCCESS), False otherwise
        """
        return True if self.state == ActionStateEnum.SUCCESS else False


class CreateOutputPackageBasesTask(Task):
    """A Task to create a list of OutputPackageBase instances from a list of Paths.

    If a pkgbase_urls dict is provided, the AnyUrl instances in it, which match the created pkgbases, will be added to
    the respective OutputPackageBase objects.

    Attributes
    ----------
    architecture: ArchitectureEnum
        A member of ArchitectureEnum that specifies the target CPU architecture that the OutputPackageBase instances
        must match
    pkgbases: list[OutputPackageBase]
        A list of OutputPackageBase instances created from the input of the task (defaults to [])
    pkgbase_urls: dict[str, AnyUrl] | None
        An optional dict, providing pkgbases and their source URLs (defaults to None)
    debug_repo: bool
        A boolean value indicating whether a debug repository is targetted
    """

    def __init__(
        self,
        architecture: ArchitectureEnum,
        package_paths: list[Path],
        with_signature: bool,
        debug_repo: bool,
        pkgbase_urls: dict[str, AnyUrl] | None = None,
        package_verification: PkgVerificationTypeEnum | None = None,
        dependencies: list[Task] | None = None,
    ):
        """Initialize an instance of CreateOutputPackageBasesTask.

        Parameters
        ----------
        architecture: ArchitectureEnum
            A member of ArchitectureEnum that specifies the target CPU architecture that the OutputPackageBase instances
            must match
        package_paths: list[Path]
            The path to a package file
        with_signature: bool
            A boolean value indicating whether to also consider signature files
        debug_repo: bool
            A boolean value indicating whether a debug repository is targetted
        pkgbase_urls: dict[str, AnyUrl] | None
            An optional dict, providing pkgbases and their source URLs (defaults to None)
        package_verification: PkgVerificationTypeEnum | None
            The type of package verification to be run against the package (defaults to None)
        dependencies: list[Task] | None
            An optional list of Task instances that are run before this task (defaults to None)
        """
        pre_checks: list[Check] = []
        post_checks: list[Check] = []

        debug(f"Initializing Task to create instances of OutputPackageBase using paths {package_paths}...")

        self.architecture = architecture
        self.debug_repo = debug_repo

        if dependencies is not None:
            self.dependencies = dependencies

        if with_signature:
            self.package_paths = [[package_path, Path(str(package_path) + ".sig")] for package_path in package_paths]
        else:
            self.package_paths = [[package_path] for package_path in package_paths]

        if with_signature:
            if package_verification == PkgVerificationTypeEnum.PACMANKEY:
                debug(f"Adding pacman-key based verification of packages {self.package_paths} to Task...")
                pre_checks.append(PacmanKeyPackagesSignatureVerificationCheck(packages=self.package_paths))

        self.pkgbase_urls = pkgbase_urls or {}
        self.pre_checks = pre_checks
        self.post_checks = post_checks
        self.pkgbases: list[OutputPackageBase] = []

    def do(self) -> ActionStateEnum:
        """Create instances of OutputPackageBase.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise.
        """
        packages: list[Package] = []

        debug(f"Running Task to create a list of OutputPackageBase instances using {self.package_paths}...")
        self.state = ActionStateEnum.STARTED_TASK

        packages_and_paths: list[tuple[Package, Path]] = []
        for package_list in self.package_paths:
            try:
                package = asyncio.run(
                    Package.from_file(
                        package=package_list[0],
                        signature=package_list[1] if len(package_list) == 2 else None,
                    )
                )
                packages.append(package)
                packages_and_paths.append((package, package_list[0]))
            except RepoManagementFileError as e:
                info(e)
                self.state = ActionStateEnum.FAILED_TASK
                return self.state

        for key, group in groupby(packages, attrgetter("pkginfo.base")):
            debug(f"Create OutputPackageBase representing pkgbase {key}")
            try:
                outputpackagebase = OutputPackageBase.from_package(packages=list(group))
                outputpackagebase.source_url = self.pkgbase_urls.get(  # type: ignore[attr-defined]
                    outputpackagebase.base,  # type: ignore[attr-defined]
                )
                self.pkgbases.append(outputpackagebase)
            except (ValueError, RuntimeError) as e:
                info(e)
                self.state = ActionStateEnum.FAILED_TASK
                return self.state

        self.post_checks.append(DebugPackagesCheck(packages=packages, debug=self.debug_repo))
        self.post_checks.append(MatchingArchitectureCheck(architecture=self.architecture, packages=packages))
        self.post_checks.append(MatchingFilenameCheck(packages_and_paths=packages_and_paths))

        self.state = ActionStateEnum.SUCCESS_TASK
        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo the creation of OutputPackageBase instances.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        debug(f"Undoing Task to create OutputPackageBase instances from packages {self.package_paths}... ")

        self.pkgbases.clear()
        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state


class PrintOutputPackageBasesTask(Task):
    """A Task to print instances of OutputPackageBase to stdout.

    Attributes
    ----------
    dumps_option: int
        An option parameter for orjson's dumps method
    pkgbases: list[OutputPackageBase]
        A list of OutputPackageBase instances to print (defaults to [])
    dependencies: list[Task] | None
        An optional list of Task instances that are run before this task (defaults to None)
    input_from_dependency: bool
        A boolean value indicating whether the Task derives its list of OutputPackageBase instances from a dependency
        Task (defaults to False)
    """

    def __init__(
        self,
        dumps_option: int,
        pkgbases: list[OutputPackageBase] | None = None,
        dependencies: list[Task] | None = None,
    ):
        """Initialize and instance of PrintOutputPackageBasesTask.

        Parameters
        ----------
        dumps_option: int
            An option parameter for orjson's dumps method
        pkgbases: list[OutputPackageBase] | None
            An optional list of OutputPackageBase instances to print
        dependencies: list[Task] | None
            An optional list of Task instances that are run before this task (defaults to None)
        """
        self.input_from_dependency = False
        self.pkgbases: list[OutputPackageBase] = []

        if dependencies is not None:
            self.dependencies = dependencies
            for dependency in self.dependencies:
                if isinstance(dependency, CreateOutputPackageBasesTask):
                    self.input_from_dependency = True

        debug("Initialize Task to print OutputPackageBase instances...")

        if not self.input_from_dependency:
            if pkgbases is None:
                raise RuntimeError("Pkgbases must be provided if not deriving input from another Task!")

            self.pkgbases = pkgbases

        self.dumps_option = dumps_option

    def do(self) -> ActionStateEnum:
        """Print OutputPackageBase instances in JSON representation.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise.
        """
        if self.input_from_dependency:
            for dependency in [
                dependency for dependency in self.dependencies if isinstance(dependency, CreateOutputPackageBasesTask)
            ]:
                match dependency.state:
                    case ActionStateEnum.SUCCESS:
                        self.pkgbases += dependency.pkgbases
                    case _:
                        self.state = ActionStateEnum.FAILED_DEPENDENCY
                        return self.state

        debug("Running Task to print OutputPackageBases in JSON format...")
        self.state = ActionStateEnum.STARTED_TASK

        for outputpackagebase in self.pkgbases:
            try:
                print(dumps(outputpackagebase.model_dump(mode="json"), option=self.dumps_option).decode("utf-8"))
            except JSONEncodeError as e:
                info(e)
                return ActionStateEnum.FAILED_TASK

        self.state = ActionStateEnum.SUCCESS_TASK
        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo the printing of OutputPackageBase instances.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        if self.state == ActionStateEnum.NOT_STARTED:
            debug("Undoing Task to print OutputPackageBase instances not possible, as it never ran...")
            return self.state

        debug("Undoing Task to print OutputPackageBase instances...")

        if self.input_from_dependency:
            self.pkgbases.clear()

        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state


class WriteOutputPackageBasesToTmpFileInDirTask(Task):
    """A Task to write instances of OutputPackageBase to temporary JSON files in a directory.

    Attributes
    ----------
    filenames: list[Path] | None
        An optional list of Paths representing the OutputPackageBase instances that are written
    pkgbases: list[OutputPackageBase]
        A list of OutputPackageBase instances to write to file
    directory: Path
        A directory Path to write the files to
    """

    def __init__(
        self,
        directory: Path,
        dumps_option: int = ORJSON_OPTION,
        pkgbases: list[OutputPackageBase] | None = None,
        dependencies: list[Task] | None = None,
    ):
        """Initialize and instance of WriteOutputPackageBasesToTmpFileInDirTask.

        Parameters
        ----------
        directory: Path
            A directory Path to write the files to
        dumps_option: int
            An option parameter for orjson's dumps method (defaults to repod.config.defaults.ORJSON_OPTION)
        pkgbases: list[OutputPackageBase] | None
            A list of OutputPackageBase instances to write to files
        dependencies: list[Task] | None
            An optional list of Task instances that are run before this task (defaults to None)
        """
        self.pkgbases = []
        self.input_from_dependency = False

        if dependencies is not None:
            self.dependencies = dependencies
            for dependency in self.dependencies:
                if isinstance(dependency, CreateOutputPackageBasesTask):
                    self.input_from_dependency = True

        self.filenames: list[Path] = []
        self.directory = directory
        self.dumps_option = dumps_option

        if self.input_from_dependency:
            debug(
                "Creating Task to write instances of OutputPackageBase to a directory, "
                "using output from another Task..."
            )
        else:
            if pkgbases is None:
                raise RuntimeError("Pkgbases are required arguments, when not depending on another Task for input!")

            debug("Creating Task to write instances of OutputPackageBase to a directory...")
            self.pkgbases = pkgbases

    def do(self) -> ActionStateEnum:
        """Write instances of OutputPackageBase to temporary JSON files in a directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise.
        """
        pkgname_dir = self.directory / "pkgnames"

        self.state = ActionStateEnum.STARTED_TASK

        if self.input_from_dependency:
            debug(
                "Running Task to write OutputPackageBase instances to a management repository directory, "
                "using output from another Task..."
            )
            for dependency in self.dependencies:
                if isinstance(dependency, CreateOutputPackageBasesTask):
                    match dependency.state:
                        case ActionStateEnum.SUCCESS:
                            self.pkgbases = dependency.pkgbases
                        case _:
                            self.state = ActionStateEnum.FAILED_DEPENDENCY
                            return self.state
        else:
            debug("Running Task to write OutputPackageBase instances to a management repository directory...")

        pkgname_dir.mkdir(parents=True, exist_ok=True)

        for outputpackagebase in self.pkgbases:
            filename = self.directory / Path(f"{outputpackagebase.base}.json.tmp")  # type: ignore[attr-defined]
            self.filenames.append(filename)

            try:
                with open(filename, "wb") as output_file:
                    output_file.write(dumps(outputpackagebase.model_dump(mode="json"), option=self.dumps_option))
            except (OSError, BlockingIOError, JSONEncodeError) as e:
                info(e)
                self.state = ActionStateEnum.FAILED_TASK
                return self.state

            target = self.directory / Path(f"{outputpackagebase.base}.json")  # type: ignore[attr-defined]
            for pkg in outputpackagebase.packages:  # type: ignore[attr-defined]
                symlink_path = pkgname_dir / Path(pkg.name + ".json.tmp")
                self.filenames.append(symlink_path)
                symlink_path.symlink_to(relative_to_shared_base(path_b=symlink_path, path_a=target))

        self.state = ActionStateEnum.SUCCESS_TASK
        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo the writing of OutputPackageBase instances to temporary JSON files in a directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        debug(
            "Undoing Task to create an OutputPackageBase instance from a list of packages and "
            "writing it to a management repository directory..."
        )

        if self.state == ActionStateEnum.NOT_STARTED:
            info("Can not undo writing of OutputPackageBase to directory {self.directory}, as it never took place.")
            return self.state

        for filename in self.filenames:
            filename.unlink(missing_ok=True)
        self.filenames.clear()

        if self.input_from_dependency:
            self.pkgbases.clear()

        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state


class MoveTmpFilesTask(Task):
    """A Task to move temporary files.

    A backup of the destination (if it exists) is created prior to moving a file

    Attributes
    ----------
    paths: list[SourceDestination]
        A list of SourceDestination instances which represent the source and destination (plus additional data) for
        each file to be moved
    input_from_dependency: bool
        A boolean value indicating whether input is derived from a dependency Task (defaults to False)
    dependencies: list[Task] | None
        An optional list of Task instances that are run before this task (defaults to None)
    """

    def __init__(
        self,
        paths: list[list[Path]] | None = None,
        dependencies: list[Task] | None = None,
    ):
        """Initialize an instance of MoveTmpFilesTask.

        If a WriteOutputPackageBasesToTmpFileInDirTask or WriteSyncDbsToTmpFilesInDirTask is provided as dependency
        Task, paths is derived from it (the input ordering of the Tasks is honored and the first match is used), else
        paths must be provided.

        Parameters
        ----------
        paths: list[list[Path]] | None
            An optional list of Path lists which represent the source and destination for each file to be moved
        dependencies: list[Task] | None
            An optional list of Task instances that are run before this task (defaults to None)
        """
        self.paths = []
        self.input_from_dependency = False

        if dependencies is not None:
            self.dependencies = dependencies
            for dependency in self.dependencies:
                if isinstance(dependency, (WriteOutputPackageBasesToTmpFileInDirTask, WriteSyncDbsToTmpFilesInDirTask)):
                    self.input_from_dependency = True

        if self.input_from_dependency:
            debug("Creating Task to move a temporary file to a destination, using output from another Task...")
        else:
            if not paths:
                raise RuntimeError(
                    "A list of Path lists is a required argument, when not depending on another Task for input!"
                )
            if not all(len(path_list) == 2 for path_list in paths):
                raise RuntimeError("Path lists are required to be supplied in lists of length two!")

            debug(f"Creating Task to move {paths}...")
            self.paths = [
                SourceDestination(
                    source=path_list[0],
                    destination=path_list[1],
                    destination_backup=Path(str(path_list[1]) + ".bkp"),
                )
                for path_list in paths
            ]

    def do(self) -> ActionStateEnum:
        """Move files from their source to their destination (with potential backup of destination).

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise.
        """
        if self.input_from_dependency and len(self.dependencies) > 0:
            debug("Getting temporary files and their destinations from the output of another Task...")
            # NOTE: we can skip branch coverage here, as the only way to go through an empty list of dependencies is by
            # manipulating MoveTmpFilesTask.input_from_dependency after initialization
            for dependency in self.dependencies:  # pragma: no branch
                if isinstance(dependency, (WriteSyncDbsToTmpFilesInDirTask, WriteOutputPackageBasesToTmpFileInDirTask)):
                    match dependency.state:
                        case ActionStateEnum.SUCCESS if isinstance(
                            dependency, WriteOutputPackageBasesToTmpFileInDirTask
                        ):
                            try:
                                self.paths = [
                                    SourceDestination(
                                        source=filename,
                                        destination=Path(str(filename).replace(".tmp", "")),
                                        destination_backup=(Path(str(filename).replace(".tmp", "") + ".bkp")),
                                    )
                                    for filename in dependency.filenames
                                ]
                            except ValidationError as e:
                                info(e)
                                self.state = ActionStateEnum.FAILED_TASK
                                return self.state
                            break
                        case ActionStateEnum.SUCCESS if isinstance(dependency, WriteSyncDbsToTmpFilesInDirTask):
                            try:
                                self.paths = [
                                    SourceDestination(
                                        source=filename,
                                        destination=Path(str(filename).replace(".tmp", "")),
                                        destination_backup=(Path(str(filename).replace(".tmp", "") + ".bkp")),
                                    )
                                    for filename in [
                                        dependency.default_syncdb_path,
                                        dependency.default_syncdb_symlink_path,
                                        dependency.files_syncdb_path,
                                        dependency.files_syncdb_symlink_path,
                                    ]
                                ]
                            except ValidationError as e:
                                info(e)
                                self.state = ActionStateEnum.FAILED_TASK
                                return self.state
                            break
                        case _:
                            self.state = ActionStateEnum.FAILED_DEPENDENCY
                            return self.state

        debug(f"Running Task to move {self.paths}...")

        self.state = ActionStateEnum.STARTED_TASK

        for source_destination in self.paths:
            if source_destination.destination.exists():
                debug(f"Backing up {source_destination.destination} to {source_destination.destination_backup}...")
                try:
                    copy2(src=source_destination.destination, dst=source_destination.destination_backup)
                except Exception as e:
                    info(e)
                    self.state = ActionStateEnum.FAILED_TASK
                    return self.state
                source_destination.backup_done = True

            try:
                source_destination.source.rename(source_destination.destination)
            except Exception as e:
                info(e)
                self.state = ActionStateEnum.FAILED_TASK
                return self.state

        self.state = ActionStateEnum.SUCCESS_TASK
        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo the moving of a file from source to destination.

        Undo also all Tasks that have been run as a dependency.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        if self.state in [ActionStateEnum.NOT_STARTED, ActionStateEnum.FAILED_DEPENDENCY]:
            info(f"Can not undo moving of files {self.paths}, as it never took place.")
            self.state = ActionStateEnum.NOT_STARTED
            self.dependency_undo()
            return self.state

        for source_destination in self.paths:
            match (
                self.state,
                source_destination.source.exists(),
                source_destination.destination.exists(),
                source_destination.backup_done,
                source_destination.destination_backup.exists(),
            ):
                case (ActionStateEnum.SUCCESS_TASK, False, True, True, True) | (
                    ActionStateEnum.SUCCESS,
                    False,
                    True,
                    True,
                    True,
                ):
                    debug(f"Moving {source_destination.destination} back to {source_destination.source}...")
                    source_destination.destination.rename(source_destination.source)
                    debug(f"Moving {source_destination.destination_backup} back to {source_destination.destination}...")
                    source_destination.destination_backup.rename(source_destination.destination)
                    self.state = ActionStateEnum.NOT_STARTED
                case (ActionStateEnum.SUCCESS_TASK, False, True, False, False) | (
                    ActionStateEnum.SUCCESS,
                    False,
                    True,
                    False,
                    False,
                ):
                    debug(f"Moving {source_destination.destination} back to {source_destination.source}...")
                    source_destination.destination.rename(source_destination.source)
                    self.state = ActionStateEnum.NOT_STARTED
                case (ActionStateEnum.FAILED_TASK, True, False, False, False):
                    self.state = ActionStateEnum.NOT_STARTED
                case (ActionStateEnum.FAILED_TASK, True, True, True, True):
                    debug(
                        f"Removing backup {source_destination.destination_backup} of "
                        f" destination {source_destination.destination}..."
                    )
                    source_destination.destination_backup.unlink()
                    self.state = ActionStateEnum.NOT_STARTED
                case _:  # pragma: no cover
                    info(f"Can not undo moving of files {self.paths}!")
                    self.state = ActionStateEnum.FAILED_UNDO_TASK

        self.dependency_undo()
        return self.state


class FilesToRepoDirTask(Task):
    """A Task to copy files to a package pool directory and create symlinks for them in package repository directories.

    Attributes
    ----------
    files: list[Path]
        A list of files to copy and create symlinks for
    file_type: RepoFileEnum
        An instance of RepoFileEnum, indicating what type of RepoFile to initialize
    settings: UserSettings | SystemSettings
        A instance of Settings to derive package_repo_dir and package_pool_dir from
    name: Path
        The name of a repository
    architecture: ArchitectureEnum | None
        The optional architecture of the target repository
    repo_type: RepoTypeEnum
        A member of RepoTypeEnum, which indicates which type of repository is targeted
    repo_files: list[RepoFile]
        A a list of RepoFile instances that represent the files and their targets (defaults to [])
    """

    def __init__(
        self,
        files: list[Path],
        file_type: RepoFileEnum,
        settings: UserSettings | SystemSettings,
        name: Path,
        architecture: ArchitectureEnum | None,
        repo_type: RepoTypeEnum,
        dependencies: list[Task] | None = None,
    ):
        """Initialize an instance of FilesToRepoDirTask.

        Parameters
        ----------
        files: list[Path]
            A list of files to copy and create symlinks for
        file_type: RepoFileEnum
            An instance of RepoFileEnum, indicating what type of RepoFile to initialize
        settings: UserSettings | SystemSettings
            A instance of Settings to derive package_repo_dir and package_pool_dir from
        name: Path
            The name of a repository
        architecture: ArchitectureEnum | None
            The optional architecture of the target repository
        repo_type: RepoTypeEnum
            A member of RepoTypeEnum, which indicates which type of repository is targeted
        dependencies: list[Task] | None
            An optional list of Task instances that are run before this task (defaults to None)
        """
        debug(f"Creating Task to move {files} to repo {name} ({architecture})...")

        if dependencies is not None:
            self.dependencies = dependencies

        self.files = files
        self.file_type = file_type
        self.settings = settings
        self.name = name
        self.architecture = architecture
        self.repo_type = repo_type
        self.repo_files: list[RepoFile] = []

    def do(self) -> ActionStateEnum:
        """Copy files to a package pool directory and create symlinks for them in a package repository directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise.
        """
        self.state = ActionStateEnum.STARTED_TASK

        debug(f"Running Task to move {self.files} to repo {self.name} ({self.architecture})...")

        try:
            package_repo_dir = self.settings.get_repo_path(
                repo_dir_type=RepoDirTypeEnum.PACKAGE,
                name=self.name,
                architecture=self.architecture,
                repo_type=self.repo_type,
            )
            package_pool_dir = self.settings.get_repo_path(
                repo_dir_type=RepoDirTypeEnum.POOL,
                name=self.name,
                architecture=self.architecture,
                repo_type=self.repo_type,
            )
        except RuntimeError as e:
            info(e)
            self.state = ActionStateEnum.FAILED_TASK
            return self.state

        for file_path in self.files:
            try:
                repo_file = RepoFile(
                    file_type=self.file_type,
                    file_path=package_pool_dir / file_path.name,
                    symlink_path=package_repo_dir / file_path.name,
                )
                self.repo_files.append(repo_file)
            except (ValidationError, RuntimeError) as e:
                info(e)
                self.state = ActionStateEnum.FAILED_TASK
                return self.state

            try:
                repo_file.copy_from(path=file_path)
                repo_file.link()
            except RepoManagementFileError as e:
                info(e)
                self.state = ActionStateEnum.FAILED_TASK
                return self.state

        self.state = ActionStateEnum.SUCCESS_TASK
        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo copying files to a package pool directory and creating symlinks in a package repository directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        if self.state == ActionStateEnum.NOT_STARTED:
            info(
                f"Can not undo moving of packages {self.files} to repository {self.name} ({self.architecture}) "
                f"as it has not happened yet!"
            )
            self.dependency_undo()
            return self.state

        for repo_file in self.repo_files:
            repo_file.remove(force=True, unlink=True)
        self.repo_files.clear()

        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state


class AddToRepoTask(Task):
    """Add package files to a repository.

    Attributes
    ----------
    dependencies: list[Task]
        A list of Tasks that are dependencies of this one
    """

    def __init__(self, dependencies: list[Task]):
        """Initialize an instance of AddToRepoTask.

        Parameters
        ----------
        dependencies: list[Task]
            A list of Tasks that are dependencies of this one
        """
        self.dependencies = dependencies

    def do(self) -> ActionStateEnum:
        """Run Task to add package files to a repository.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise
        """
        self.state = ActionStateEnum.SUCCESS_TASK
        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo the adding of packages to a repository.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state


class WriteSyncDbsToTmpFilesInDirTask(Task):
    """A Task to write a repository's sync databases and their symlinks to temporary files in a directory.

    Attributes
    ----------
    compression: CompressionTypeEnum
        A member of CompressionTypeEnum which is used as compression type for the repository sync databases
    desc_version: PackageDescVersionEnum
        A member of PackageDescVersionEnum which is used to set the PackageDesc version to write to file
    files_version: FilesVersionEnum
        A member of FilesVersionEnum which is used to set the PackageDesc version to write to file
    management_repo_dir: Path
        A Path to a directory in a management repository from which to read JSON files
    default_syncdb_path: Path
        A Path for the temporary default repository sync database
    files_syncdb_path: Path
        A Path for the temporary files repository sync database
    default_syncdb_symlink_path: Path
        A Path for the temporary symlink to the default repository sync database
    files_syncdb_symlink_path: Path
        A Path for the temporary symlink to the files repository sync database
    dependencies: list[Task] | None
        An optional list of Task lists which are executed before this Task (defaults to None)
    """

    def __init__(
        self,
        compression: CompressionTypeEnum,
        desc_version: PackageDescVersionEnum,
        files_version: FilesVersionEnum,
        management_repo_dir: Path,
        package_repo_dir: Path,
        dependencies: list[Task] | None = None,
    ):
        """Initialize an instance of WriteSyncDbsToTmpFilesInDirTask.

        Parameters
        ----------
        compression: CompressionTypeEnum
            A member of CompressionTypeEnum which is used as compression type for the repository sync databases
        desc_version: PackageDescVersionEnum
            A member of PackageDescVersionEnum which is used to set the PackageDesc version to write to file
        files_version: FilesVersionEnum
            A member of FilesVersionEnum which is used to set the PackageDesc version to write to file
        management_repo_dir: Path
            A Path to a directory in a management repository from which to read JSON files
        package_repo_dir: Path
            A Path to a directory in a package repository to write files to
        dependencies: list[Task] | None
            An optional list of Task lists which are executed before this Task (defaults to None)
        """
        self.compression = compression
        self.desc_version = desc_version
        self.files_version = files_version
        self.management_repo_dir = management_repo_dir
        self.default_syncdb_path = package_repo_dir / Path(
            package_repo_dir.parent.name + CompressionTypeEnum.db_tar_suffix(compression_type=compression) + ".tmp"
        )
        self.default_syncdb_symlink_path = package_repo_dir / Path(package_repo_dir.parent.name + ".db.tmp")
        self.files_syncdb_path = package_repo_dir / Path(
            package_repo_dir.parent.name
            + CompressionTypeEnum.db_tar_suffix(compression_type=compression, files=True)
            + ".tmp"
        )
        self.files_syncdb_symlink_path = package_repo_dir / Path(package_repo_dir.parent.name + ".files.tmp")
        if dependencies:
            self.dependencies = dependencies

    def do(self) -> ActionStateEnum:
        """Run Task to write temporary repository sync databases to a package repository directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise
        """
        self.state = ActionStateEnum.STARTED_TASK

        debug(
            f"Running Task to write temporary repository sync databases {self.default_syncdb_path} and "
            f"{self.files_syncdb_path}..."
        )

        try:
            default_sync_db = SyncDatabase(
                database=self.default_syncdb_path,
                database_type=RepoDbTypeEnum.DEFAULT,
                compression_type=self.compression,
                desc_version=self.desc_version,
                files_version=self.files_version,
            )
            files_sync_db = SyncDatabase(
                database=self.files_syncdb_path,
                database_type=RepoDbTypeEnum.FILES,
                compression_type=self.compression,
                desc_version=self.desc_version,
                files_version=self.files_version,
            )
        except ValidationError as e:
            info(e)
            self.state = ActionStateEnum.FAILED_TASK
            return self.state

        try:
            asyncio.run(default_sync_db.stream_management_repo(path=self.management_repo_dir))
            asyncio.run(files_sync_db.stream_management_repo(path=self.management_repo_dir))
        except (IsADirectoryError, RepoManagementFileNotFoundError) as e:
            info(e)
            self.state = ActionStateEnum.FAILED_TASK
            return self.state

        self.default_syncdb_symlink_path.symlink_to(
            Path(sub(r"\.tmp$", "", str(self.default_syncdb_path))).relative_to(self.default_syncdb_symlink_path.parent)
        )
        self.files_syncdb_symlink_path.symlink_to(
            Path(sub(r"\.tmp$", "", str(self.files_syncdb_path))).relative_to(self.files_syncdb_symlink_path.parent)
        )

        self.state = ActionStateEnum.SUCCESS_TASK
        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo the writing of temporary repository sync databases in a package repository directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        if self.state == ActionStateEnum.NOT_STARTED:
            info(
                f"Can not undo writing of temporary repository sync databases {self.default_syncdb_path} and "
                f"{self.files_syncdb_path}, as it has not happened yet!"
            )
            self.dependency_undo()
            return self.state

        for path in [
            self.files_syncdb_path,
            self.files_syncdb_symlink_path,
            self.default_syncdb_path,
            self.default_syncdb_symlink_path,
        ]:
            if not path.is_dir():
                path.unlink(missing_ok=True)

        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state


class RemoveBackupFilesTask(Task):
    """A destructive(!) Task to remove backup files.

    Attributes
    ----------
    paths: list[Path]
        A list of backup file paths to remove
    dependencies: list[Task] | None
        An optional list of Task lists which are executed before this Task (defaults to None)
    """

    def __init__(self, paths: list[Path] | None = None, dependencies: list[Task] | None = None):
        """Initialize an instance of RemoveBackupFilesTask.

        If instances of MoveTmpFilesTask are provided in dependencies, paths is populated from them.

        Parameters
        ----------
        paths: list[Path] | None
            An optional list of Paths which represent the backup files to be removed
        dependencies: list[Task] | None
            An optional list of Task instances that are run before this task (defaults to None)
        """
        self.input_from_dependency: bool = False
        if dependencies:
            self.dependencies = dependencies
            for dependency in self.dependencies:
                if isinstance(dependency, MoveTmpFilesTask):
                    self.input_from_dependency = True

        if self.input_from_dependency:
            debug("Creating Task to remove backup files, using output from another Task...")
            self.paths = []
        else:
            if not paths:
                raise RuntimeError("Paths must be provided if not depending on another Task for input!")

            debug(f"Creating Task to backup files {paths}...")
            self.paths = paths

    def do(self) -> ActionStateEnum:
        """Run Task to remove backup files.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise
        """
        if self.input_from_dependency and len(self.dependencies) > 0:
            debug("Getting backup files from the output of another Task...")
            for dependency in self.dependencies:  # pragma: no branch
                if isinstance(dependency, MoveTmpFilesTask):
                    if dependency.state == ActionStateEnum.SUCCESS:
                        self.paths += [obj.destination_backup for obj in dependency.paths]
                    else:
                        self.state = ActionStateEnum.FAILED_DEPENDENCY
                        return self.state

        debug(f"Running Task to remove backup files {self.paths}...")
        self.state = ActionStateEnum.STARTED_TASK

        for path in self.paths:
            path.unlink(missing_ok=True)

        self.state = ActionStateEnum.SUCCESS_TASK

        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo Task for the removal of backup files.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        if self.state == ActionStateEnum.NOT_STARTED:
            info(f"Can not undo removing of backup files {self.paths} as it has not happened yet!")
            self.dependency_undo()
            return self.state

        if self.input_from_dependency:
            self.paths.clear()

        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state


class ConsolidateOutputPackageBasesTask(Task):
    """Task to compare OutputPackageBase instances with those in a management repository.

    Attributes
    ----------
    pkgbases: list[OutputPackageBase]
        A list of OutputPackageBase instances to compare to those in a management repository directory
    directory: Path
        A Path to the directory in a management repository where to look for OutputPackageBases
    filenames: list[str]
        A list of package filenames associated with the OutputPackageBases in pkgbases
    package_names: list[str]
        A list of package names associated with the OutputPackageBases in pkgbases
    current_filenames: list[str]
        A list of package filenames associated with the pkgbases in directory, that match those in pkgbases
    current_package_names: list[str]
        A list of package names associated with the pkgbases in directory, that match those in pkgbases
    url_validation_settings: UrlValidationSettings | None
        An optional instance of UrlValidationSettings providing settings for validating the source URLs of pkgbases
        (defaults to None)
    """

    def __init__(
        self,
        directory: Path,
        stability_layer_dirs: tuple[list[Path], list[Path]],
        url_validation_settings: UrlValidationSettings | None = None,
        pkgbases: list[OutputPackageBase] | None = None,
        dependencies: list[Task] | None = None,
    ):
        """Initialize an instance of ConsolidateOutputPackageBasesTask.

        If instances of CreateOutputPackageBasesTask are provided in dependencies, pkgbases is populated from them.

        Parameters
        ----------
        directory: Path
            A Path to the directory in a management repository where to look for OutputPackageBases
        stability_layer_dirs: tuple[list[Path], list[Path]]
            A tuple of two Path lists that represent the stability layers above (first list) and below (second list) the
            current
        pkgbases: list[OutputPackageBase] | None
            An optional list of OutputPackageBase instances to compare to those in a management repository directory
        url_validation_settings: UrlValidationSettings | None
            An optional instance of UrlValidationSettings providing settings for validating the source URLs of pkgbases
            (defaults to None)
        dependencies: list[Task] | None
            An optional list of Task instances that are run before this task (defaults to None)
        """
        self.directory = directory
        if not self.directory or not self.directory.exists():
            raise RuntimeError("The provided directory must exist!")

        self.stability_layer_dirs = stability_layer_dirs

        self.url_validation_settings = url_validation_settings
        self.input_from_dependency = False

        if dependencies:
            self.dependencies = dependencies
            for dependency in self.dependencies:
                if isinstance(dependency, CreateOutputPackageBasesTask):
                    self.input_from_dependency = True

        if self.input_from_dependency:
            debug("Creating Task to compare OutputPackageBases, using output from another Task...")
            self.pkgbases = []
        else:
            if not pkgbases:
                raise RuntimeError("Pkgbases must be provided if not depending on another Task for input!")

            debug(
                "Creating Task to compare pkgbases "
                f"{[pkgbase.base for pkgbase in pkgbases]}..."  # type: ignore[attr-defined]
            )
            self.pkgbases = pkgbases

        self.filenames: list[str] = []
        self.package_names: list[str] = []
        self.current_filenames: list[str] = []
        self.current_package_names: list[str] = []

    def do(self) -> ActionStateEnum:
        """Run Task to compare OutputPackageBase instances with those in a management repository directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise
        """
        if self.input_from_dependency and len(self.dependencies) > 0:
            debug("Getting pkgbases from the output of another Task...")
            for dependency in self.dependencies:  # pragma: no branch
                if isinstance(dependency, CreateOutputPackageBasesTask):
                    if dependency.state == ActionStateEnum.SUCCESS:
                        self.pkgbases = dependency.pkgbases
                    else:
                        self.state = ActionStateEnum.FAILED_DEPENDENCY
                        return self.state

        pkgbase_names = [pkgbase.base for pkgbase in self.pkgbases]  # type: ignore[attr-defined]
        self.filenames = [
            package.filename
            for package_list in [pkgbase.packages for pkgbase in self.pkgbases]  # type: ignore[attr-defined]
            for package in package_list
        ]
        self.package_names = [
            package.name
            for package_list in [pkgbase.packages for pkgbase in self.pkgbases]  # type: ignore[attr-defined]
            for package in package_list
        ]

        debug(f"Running Task to consolidate pkgbases {pkgbase_names}...")
        self.state = ActionStateEnum.STARTED_TASK

        current_pkgbases: list[OutputPackageBase] = []
        pkgbases_above: list[OutputPackageBase] = []
        pkgbases_below: list[OutputPackageBase] = []

        try:
            read_pkgbases_from_stability_layers(
                directory=self.directory,
                pkgbase_names=pkgbase_names,
                stability_layer_dirs=self.stability_layer_dirs,
                current_pkgbases=current_pkgbases,
                current_filenames=self.current_filenames,
                current_package_names=self.current_package_names,
                pkgbases_above=pkgbases_above,
                pkgbases_below=pkgbases_below,
            )
        except RepoManagementFileError as e:
            info(e)
            self.state = ActionStateEnum.FAILED_TASK
            return self.state

        debug(f"Found new package filenames: {self.filenames}")
        debug(f"Found new package names: {self.package_names}")
        debug(f"Current package filenames: {self.current_filenames}")
        debug(f"Current package names: {self.current_package_names}")

        self.post_checks.append(
            StabilityLayerCheck(
                pkgbases=self.pkgbases,
                pkgbases_above=pkgbases_above,
                pkgbases_below=pkgbases_below,
            )
        )
        self.post_checks.append(
            SourceUrlCheck(
                new_pkgbases=self.pkgbases,
                current_pkgbases=current_pkgbases,
                url_validation_settings=self.url_validation_settings,
            )
        )
        self.post_checks.append(
            PkgbasesVersionUpdateCheck(new_pkgbases=self.pkgbases, current_pkgbases=current_pkgbases),
        )
        self.post_checks.append(
            PackagesNewOrUpdatedCheck(
                directory=self.directory,
                new_pkgbases=self.pkgbases,
                current_pkgbases=current_pkgbases,
            ),
        )

        self.state = ActionStateEnum.SUCCESS_TASK

        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo Task to consolidate OutputPackageBase instances with those from a management repository directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        if self.state == ActionStateEnum.NOT_STARTED:
            info(
                "Can not undo consolidation of OutputPackageBases "
                f"{[pkgbase.base for pkgbase in self.pkgbases]} "  # type: ignore[attr-defined]
                "as it has not happened yet!"
            )
            self.dependency_undo()
            return self.state

        if self.input_from_dependency:
            self.pkgbases.clear()

        self.filenames.clear()
        self.package_names.clear()
        self.current_filenames.clear()
        self.current_package_names.clear()

        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state


class RemoveManagementRepoSymlinksTask(Task):
    """Destructive Task to remove symlinks in a management repository directory.

    Attributes
    ----------
    directory: Path
        A Path to the directory in a management repository
    names: list[str]
        A list of names, representing package in the management repository
    """

    def __init__(self, directory: Path, names: list[str] | None = None, dependencies: list[Task] | None = None):
        """Initialize an instance of RemoveManagementRepoSymlinksTask.

        If instances of ConsolidateOutputPackageBasesTask are provided in dependencies, names is populated from them.

        Parameters
        ----------
        directory: Path
            A Path to the directory in a management repository
        names: list[str] | None
            An aoptional list of names, representing package in the management repository (defaults to None)
        dependencies: list[Task] | None
            An optional list of Task instances that are run before this task (defaults to None)
        """
        self.directory = directory
        if not self.directory or not self.directory.exists():
            raise RuntimeError("The provided directory must exist!")

        self.input_from_dependency = False

        if dependencies:
            self.dependencies = dependencies
            for dependency in self.dependencies:
                if isinstance(dependency, ConsolidateOutputPackageBasesTask):
                    self.input_from_dependency = True

        if self.input_from_dependency:
            debug(
                "Creating Task to remove symlinks in a management repository directory, "
                "using output from another Task..."
            )
            self.names = []
        else:
            if not names:
                raise RuntimeError("Names must be provided if not depending on another Task for input!")

            debug(
                f"Creating Task to remove symlinks {', '.join([name for name in names])} "
                f"in repository directory {directory}..."
            )
            self.names = names

    def do(self) -> ActionStateEnum:
        """Run Task to remove symlinks in a management repository directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise
        """
        if self.input_from_dependency and len(self.dependencies) > 0:
            debug("Getting pkgbases from the output of another Task...")
            for dependency in self.dependencies:  # pragma: no branch
                if isinstance(dependency, ConsolidateOutputPackageBasesTask):
                    if dependency.state == ActionStateEnum.SUCCESS:
                        self.names = [
                            name for name in dependency.current_package_names if name not in dependency.package_names
                        ]
                    else:
                        self.state = ActionStateEnum.FAILED_DEPENDENCY
                        return self.state

        for name in self.names:
            symlink = self.directory / f"pkgnames/{name}.json"
            debug(f"Removing symlink {symlink}...")
            symlink.unlink(missing_ok=True)

        self.state = ActionStateEnum.SUCCESS_TASK

        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo Task to remove symlinks from a management repository directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        if self.state == ActionStateEnum.NOT_STARTED:
            info(
                "Can not undo removing of symlinks for packages "
                f"{', '.join([name for name in self.names])} "
                "as it has not happened yet!"
            )
            self.dependency_undo()
            return self.state

        if self.input_from_dependency:
            self.names.clear()

        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state


class RemovePackageRepoSymlinksTask(Task):
    """Destructive Task to remove symlinks in a package repository directory.

    This Task removes symlinks for package files and their accompanying signature files.

    Attributes
    ----------
    directory: Path
        A Path to the directory in a package repository
    filenames: list[str]
        A list of filename strings
    """

    def __init__(self, directory: Path, filenames: list[str] | None = None, dependencies: list[Task] | None = None):
        """Initialize an instance of RemovePackageRepoSymlinksTask.

        If instances of ConsolidateOutputPackageBasesTask are provided in dependencies, filenames is populated from
        them.

        Parameters
        ----------
        directory: Path
            A Path to the directory in a package repository
        filenames: list[str] | None
            An optional list of filename strings (defaults to None)
        dependencies: list[Task] | None
            An optional list of Task instances that are run before this task (defaults to None)
        """
        self.directory = directory
        if not self.directory or not self.directory.exists():
            raise RuntimeError("The provided directory must exist!")

        self.input_from_dependency = False

        if dependencies:
            self.dependencies = dependencies
            for dependency in self.dependencies:
                if isinstance(dependency, ConsolidateOutputPackageBasesTask):
                    self.input_from_dependency = True

        if self.input_from_dependency:
            debug(
                "Creating Task to remove symlinks in a package repository directory, using output from another Task..."
            )
            self.filenames = []
        else:
            if not filenames:
                raise RuntimeError("Filenames must be provided if not depending on another Task for input!")

            debug(
                f"Creating Task to remove symlinks {', '.join([filename for filename in filenames])} "
                f"in package repository directory {directory}..."
            )
            self.filenames = filenames

    def do(self) -> ActionStateEnum:
        """Run Task to remove symlinks in a package repository directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise
        """
        if self.input_from_dependency and len(self.dependencies) > 0:
            debug("Getting pkgbases from the output of another Task...")
            for dependency in self.dependencies:  # pragma: no branch
                if isinstance(dependency, ConsolidateOutputPackageBasesTask):
                    if dependency.state == ActionStateEnum.SUCCESS:
                        self.filenames = dependency.current_filenames
                    else:
                        self.state = ActionStateEnum.FAILED_DEPENDENCY
                        return self.state

        for filename in self.filenames:
            package_file = self.directory / filename
            signature_file = self.directory / f"{filename}.sig"
            debug(f"Removing symlink {package_file}...")
            package_file.unlink(missing_ok=True)
            debug(f"Removing symlink {signature_file}...")
            signature_file.unlink(missing_ok=True)

        self.state = ActionStateEnum.SUCCESS_TASK

        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo Task to remove symlinks from a package repository directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        if self.state == ActionStateEnum.NOT_STARTED:
            info(
                "Can not undo removing of symlinks "
                f"{', '.join([filename for filename in self.filenames])} "
                f"from package repository directory {str(self.directory)} as it has not happened yet!"
            )
            self.dependency_undo()
            return self.state

        if self.input_from_dependency:
            self.filenames.clear()

        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state


class CleanupRepoTask(Task):
    """Cleanup files in a repository.

    Attributes
    ----------
    dependencies: list[Task]
        A list of Tasks that are dependencies of this one
    """

    def __init__(self, dependencies: list[Task]):
        """Initialize an instance of AddToRepoTask.

        Parameters
        ----------
        dependencies: list[Task]
            A list of Tasks that are dependencies of this one
        """
        self.dependencies = dependencies

    def do(self) -> ActionStateEnum:
        """Run Task to add package files to a repository.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise
        """
        self.state = ActionStateEnum.SUCCESS_TASK
        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo the adding of packages to a repository.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state


class AddToArchiveTask(Task):
    """Add files to an archive directory.

    Attributes
    ----------
    files: list[CopySourceDestination]
        A list of CopySourceDestination that represents the sources and destinations (in the archive)
    """

    def __init__(
        self,
        archive_dir: Path,
        filenames: list[Path] | None = None,
        dependencies: list[Task] | None = None,
    ):
        """Initialize an instance of AddToArchiveTask.

        If instances of FilesToRepoDirTask are added to dependencies, the list of files is derived from them

        Parameters
        ----------
        archive_dir: Path
            An archive directory below which directory structures for files are created and files are copied to for
            archiving
        filenames: list[Path] | None
            An optional list of file Paths (defaults to None)
        dependencies: list[Task] | None
            An optional list of Task instances that are run before this task (defaults to None)

        Raises
        ------
        RuntimeError
            If not providing archive_dir
            or if not providing filenames when
        """
        if not archive_dir:
            raise RuntimeError("An archive directory must be provided!")

        self.archive_dir = archive_dir

        self.input_from_dependency = False

        if dependencies:
            self.dependencies = dependencies
            for dependency in self.dependencies:
                if isinstance(dependency, FilesToRepoDirTask):
                    self.input_from_dependency = True

        if self.input_from_dependency:
            debug(
                "Creating Task to remove symlinks in a package repository directory, using output from another Task..."
            )
            self.files: list[CopySourceDestination] = []
        else:
            if not filenames:
                raise RuntimeError("Filenames must be provided if not depending on another Task for input!")

            debug(
                f"Creating Task to archive {', '.join([str(filename) for filename in filenames])} "
                f"below {archive_dir}..."
            )
            self.files = [
                CopySourceDestination.from_archive_dir(source=filename, output_dir=archive_dir)
                for filename in filenames
            ]

    def do(self) -> ActionStateEnum:
        """Run Task to add files to an archive directory.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise
        """
        if self.input_from_dependency and len(self.dependencies) > 0:
            debug("Getting pkgbases from the output of another Task...")
            for dependency in self.dependencies:  # pragma: no branch
                if isinstance(dependency, FilesToRepoDirTask):
                    if dependency.state == ActionStateEnum.SUCCESS:
                        self.files += [
                            CopySourceDestination.from_archive_dir(source=filename, output_dir=self.archive_dir)
                            for filename in dependency.files
                        ]
                    else:
                        self.state = ActionStateEnum.FAILED_DEPENDENCY
                        return self.state

        debug(
            f"Running Task to add {', '.join([str(obj.source) for obj in self.files])} to "
            f"archive directory {str(self.archive_dir)}..."
        )
        self.state = ActionStateEnum.STARTED_TASK

        for cp_source_destination in self.files:
            cp_source_destination.copy_file()

        self.state = ActionStateEnum.SUCCESS_TASK
        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo the archiving of files.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        if self.state == ActionStateEnum.NOT_STARTED:
            info(
                f"Can not undo archiving of {', '.join([str(obj.source) for obj in self.files])} "
                f"in {str(self.archive_dir)} as it has not happened yet!"
            )
            self.dependency_undo()
            return self.state

        debug(
            f"Undoing Task to archive files {', '.join([str(obj.source) for obj in self.files])} in "
            f"{str(self.archive_dir)}..."
        )

        for cp_source_destination in self.files:
            cp_source_destination.remove_destination()

        if self.input_from_dependency:
            self.files.clear()

        self.state = ActionStateEnum.NOT_STARTED
        debug(f"before dependency undo: {self.state}")
        self.dependency_undo()
        debug(f"after dependency undo: {self.state}")
        return self.state


class ReproducibleBuildEnvironmentTask(Task):
    """Gather data on OutputPackageBase from management repo, archive (if present) and other OutputPackageBases.

    Attributes
    ----------
    archive_dir: Path | None,
        A directory below which archived files may be found
    management_directories: list[Path]
        A list of Paths in a management repository where to look for OutputPackageBases
    pkgbases: list[OutputPackageBase]
        A list of OutputPackageBase instances to compare to those in a management repository directory
    pkgs_in_archive: list[str]
        A list of pkgname-pkgver-architecture strings that represent all build requirement matches in the archive
    pkgs_in_repo: list[str]
        A list of pkgname-pkgver-architecture strings that represent all build requirement matches in the current
        repository state
    pkgs_in_transaction: list[str]
        A list of pkgname-pkgver-architecture strings that represent all build requirement matches in the current
        transaction
    """

    def __init__(
        self,
        archive_dir: Path | None,
        management_directories: list[Path],
        pkgbases: list[OutputPackageBase] | None = None,
        dependencies: list[Task] | None = None,
    ):
        """Initialize an instance of ReproducibleBuildEnvironmentTask.

        If instances of CreateOutputPackageBasesTask are provided in dependencies, pkgbases is populated from them.

        Parameters
        ----------
        archive_dir: Path | None,
            A directory below which archived files may be found
        management_directories: list[Path]
            A list of Paths in a management repository where to look for OutputPackageBases
        pkgbases: list[OutputPackageBase] | None
            An optional list of OutputPackageBase instances to compare to those in a management repository directory
        dependencies: list[Task] | None
            An optional list of Task instances that are run before this task (defaults to None)
        """
        if not management_directories:
            raise RuntimeError("At least one management repository directory must be provided!")

        if not all([directory and directory.exists() for directory in management_directories]):
            raise RuntimeError("The provided management repository directories must exist!")
        self.management_directories = management_directories

        self.archive_dir = archive_dir

        self.input_from_dependency = False

        if dependencies:
            self.dependencies = dependencies
            for dependency in self.dependencies:
                if isinstance(dependency, CreateOutputPackageBasesTask):
                    self.input_from_dependency = True
        else:
            self.dependencies = []

        if self.input_from_dependency:
            debug("Creating Task to compare OutputPackageBases, using output from another Task...")
            self.pkgbases = []
        else:
            if not pkgbases:
                raise RuntimeError("Pkgbases must be provided if not depending on another Task for input!")

            debug(
                "Creating Task to gather build requirements of pkgbases "
                f"{[pkgbase.base for pkgbase in pkgbases]}..."  # type: ignore[attr-defined]
            )
            self.pkgbases = pkgbases

        self.pkgs_in_repo: set[str] = set()
        self.pkgs_in_archive: set[str] = set()
        self.pkgs_in_transaction: set[str] = set()

    def do(self) -> ActionStateEnum:  # noqa: C901
        """Run gather data on OutputPackageBase from management repo, archive (if present) and other OutputPackageBases.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise
        """
        if self.input_from_dependency:
            debug("Getting pkgbases from the output of another Task...")
            for dependency in self.dependencies:
                if isinstance(dependency, CreateOutputPackageBasesTask):
                    if dependency.state == ActionStateEnum.SUCCESS:
                        self.pkgbases = dependency.pkgbases
                    else:
                        self.state = ActionStateEnum.FAILED_DEPENDENCY
                        return self.state

        debug("Running Task to gather build requirements for packages...")
        self.state = ActionStateEnum.STARTED_TASK

        try:
            read_build_requirements_from_management_repo_dirs(
                pkgbases=self.pkgbases,
                management_directories=self.management_directories,
                pkgs_in_repo=self.pkgs_in_repo,
            )
        except TaskError as e:
            info(e)
            self.state = ActionStateEnum.FAILED_TASK
            return self.state

        try:
            read_build_requirements_from_archive_dir(
                pkgbases=self.pkgbases,
                archive_dir=self.archive_dir,
                pkgs_in_archive=self.pkgs_in_archive,
            )
        except TaskError as e:
            info(e)
            self.state = ActionStateEnum.FAILED_TASK
            return self.state

        for pkgbase in self.pkgbases:
            for pkg in pkgbase.packages:  # type: ignore[attr-defined]
                self.pkgs_in_transaction.add(f"{pkg.name}-{pkgbase.version}-{pkg.arch}")  # type: ignore[attr-defined]

        debug(
            "Found the following build requirements in the repositories: "
            f"{', '.join([dep for dep in self.pkgs_in_repo])}"
        )
        debug(
            "Found the following build requirements in the archives: "
            f"{', '.join([dep for dep in self.pkgs_in_archive])}"
        )
        debug(
            "Found the following build requirements in the transaction: "
            f"{', '.join([dep for dep in self.pkgs_in_transaction])}"
        )

        self.post_checks.append(
            ReproducibleBuildEnvironmentCheck(
                pkgbases=self.pkgbases,
                available_requirements=self.pkgs_in_repo | self.pkgs_in_archive | self.pkgs_in_transaction,
            )
        )

        self.state = ActionStateEnum.SUCCESS_TASK

        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo gather data on OutputPackageBase from management repo, archive and other OutputPackageBases.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        if self.state == ActionStateEnum.NOT_STARTED:
            info(
                "Can not undo consolidation of OutputPackageBases "
                f"{[pkgbase.base for pkgbase in self.pkgbases]} "  # type: ignore[attr-defined]
                "as it has not happened yet!"
            )
            self.dependency_undo()
            return self.state

        debug("Undo gathering of build requirements for pkgbases...")

        if self.input_from_dependency:
            self.pkgbases.clear()

        self.pkgs_in_repo.clear()
        self.pkgs_in_archive.clear()
        self.pkgs_in_transaction.clear()

        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state


class RepoGroupTask(Task):
    """A Task to retrieve information about repositories in a group.

    Attributes
    ----------
    repositories: list[PackageRepo]
        A list of PackageRepo instances, from which all management repository directories are retrieved
    pkgbases: list[OutputPackageBase] | None
        An optional list of OutputPackageBase instances from which all pkgbase and package names are retrieved
    dependencies: list[Task] | None
        An optional list of Task instances that are run before this task (defaults to None)
    pkgbase_names: list[str]
        A list of all pkgbase names, retrieved from pkgbases
    package_names: list[str]
        A list of all package names, retrieved from pkgbases
    repo_management_dirs: dict[Path, list[Path]]
        A dict of repository name and respective management repository directory Paths
    """

    def __init__(
        self,
        repositories: list[PackageRepo],
        pkgbases: list[OutputPackageBase] | None = None,
        dependencies: list[Task] | None = None,
    ):
        """Initialize an instance of RepoGroupTask.

        If instances of CreateOutputPackageBasesTask are provided in dependencies, pkgbases is populated from them.

        Parameters
        ----------
        repositories: list[PackageRepo]
            A list of PackageRepo instances, from which all management repository directories are retrieved
        pkgbases: list[OutputPackageBase] | None
            An optional list of OutputPackageBase instances from which all pkgbase and package names are retrieved
        dependencies: list[Task] | None
            An optional list of Task instances that are run before this task (defaults to None)
        """
        self.repositories = repositories

        self.input_from_dependency = False

        if dependencies:
            self.dependencies = dependencies
            for dependency in self.dependencies:
                if isinstance(dependency, CreateOutputPackageBasesTask):
                    self.input_from_dependency = True
        else:
            self.dependencies = []

        if self.input_from_dependency:
            debug("Creating Task to compare OutputPackageBases, using output from another Task...")
            self.pkgbases = []
        else:
            if not pkgbases:
                raise RuntimeError("Pkgbases must be provided if not depending on another Task for input!")

            debug(
                "Creating Task to gather build requirements of pkgbases "
                f"{[pkgbase.base for pkgbase in pkgbases]}..."  # type: ignore[attr-defined]
            )
            self.pkgbases = pkgbases

        self.pkgbase_names: list[str] = []
        self.package_names: list[str] = []
        self.repo_management_dirs: dict[Path, list[Path]] = defaultdict(list)

    def do(self) -> ActionStateEnum:
        """Run Task to gather data related to PackageRepos in a group.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.SUCCESS_TASK if the Task ran successfully,
            ActionStateEnum.FAILED_TASK otherwise
        """
        if self.input_from_dependency:
            debug("Getting pkgbases from the output of another Task...")
            for dependency in self.dependencies:
                if isinstance(dependency, CreateOutputPackageBasesTask):
                    if dependency.state == ActionStateEnum.SUCCESS:
                        self.pkgbases = dependency.pkgbases
                    else:
                        self.state = ActionStateEnum.FAILED_DEPENDENCY
                        return self.state

        debug("Running Task to gather information about repository groups...")
        self.state = ActionStateEnum.STARTED_TASK

        self.pkgbase_names = [pkgbase.base for pkgbase in self.pkgbases]  # type: ignore[attr-defined]
        self.package_names = [
            package.name
            for package_list in [pkgbase.packages for pkgbase in self.pkgbases]  # type: ignore[attr-defined]
            for package in package_list
        ]
        for repo in self.repositories:
            self.repo_management_dirs[repo.name] = repo.get_all_management_repo_dirs()

        debug(f"Found pkgbases: {', '.join(self.pkgbase_names)}")
        debug(f"Found packages: {', '.join(self.package_names)}")
        debug(f"Found management repository directories of repositories: {self.repo_management_dirs}")

        self.post_checks.append(
            UniqueInRepoGroupCheck(
                pkgbase_names=self.pkgbase_names,
                package_names=self.package_names,
                repo_management_dirs=self.repo_management_dirs,
            )
        )

        self.state = ActionStateEnum.SUCCESS_TASK
        return self.state

    def undo(self) -> ActionStateEnum:
        """Undo Task to gather data related to PackageRepos in a group.

        Returns
        -------
        ActionStateEnum
            ActionStateEnum.NOT_STARTED if undoing the Task operation is successful,
            ActionStateEnum.FAILED_UNDO_DEPENDENCY if undoing of any of the dependency Tasks failed,
            ActionStateEnum.FAILED_UNDO_TASK otherwise
        """
        if self.state == ActionStateEnum.NOT_STARTED:
            info(
                "Can not undo gathering of repository group data for pkgbases "
                f"{[pkgbase.base for pkgbase in self.pkgbases]} "  # type: ignore[attr-defined]
                "as it has not happened yet!"
            )
            self.dependency_undo()
            return self.state

        debug("Undo gathering of repository group data...")

        if self.input_from_dependency:
            self.pkgbases.clear()

        self.pkgbase_names.clear()
        self.package_names.clear()
        self.repo_management_dirs.clear()

        self.state = ActionStateEnum.NOT_STARTED
        self.dependency_undo()
        return self.state



"""Common function and tools to work with files."""

from gzip import BadGzipFile
from gzip import open as gzip_open
from io import BytesIO, StringIO
from logging import debug
from pathlib import Path
from tarfile import ReadError, TarFile
from tarfile import open as tarfile_open
from typing import IO, Literal

import magic
from pyzstd import CParameter, ZstdDict, ZstdFile

from repod.common.enums import CompressionTypeEnum
from repod.errors import RepoManagementFileError, RepoManagementFileNotFoundError


class ZstdTarFile(TarFile):
    """A class to provide reading and writing of zstandard files using TarFile functionality."""

    def __init__(  # type: ignore[no-untyped-def]
        self,
        name: str | Path,
        mode: Literal["r", "a", "w", "x"] = "r",
        level_or_option: None | int | dict[CParameter, int] = None,
        zstd_dict: ZstdDict | None = None,
        **kwargs,
    ) -> None:
        """Initialize an instance of ZstdTarFile."""
        self.zstd_file = ZstdFile(
            filename=name,
            mode=mode,
            level_or_option=level_or_option,
            zstd_dict=zstd_dict,
        )

        try:
            super().__init__(fileobj=self.zstd_file, mode=mode, **kwargs)
        except Exception as e:
            self.zstd_file.close()
            raise RepoManagementFileError(f"An error occured while trying to open the file {name}!\n{e}")

    def close(self) -> None:
        """Close the file."""
        try:
            super().close()
        finally:
            self.zstd_file.close()


def compression_type_of_tarfile(path: Path) -> CompressionTypeEnum:
    """Retrieve the compression type of a tar file.

    Parameters
    ----------
    path: Path
        The path to a tar file

    Raises
    ------
    RepoManagementFileError
        If an unknown compression type is encountered

    Returns
    -------
    CompressionTypeEnum
        A member of CompressionTypeEnum, that reflects the compression type of tar file at path
    """
    with open(path, "rb") as f:
        file_start_bytes: bytes = f.read(2048)

    # Try and detect the instance of the libmagic shared library (loaded via
    # ctypes) used by the magic.py shipped with file.
    if hasattr(magic, "_libraries"):  # pragma: no cover
        types = magic.detect_from_content(file_start_bytes).name  # type: ignore[attr-defined]
    else:
        m = magic.Magic(keep_going=True)
        types = m.from_buffer(file_start_bytes)
    types = types.lower().strip(",")
    debug(f"Types of file {path} detected as: {types}")

    if "posix tar archive" in types:
        return CompressionTypeEnum.NONE
    elif "bzip2 compressed data" in types:
        return CompressionTypeEnum.BZIP2
    elif "gzip compressed data" in types:
        return CompressionTypeEnum.GZIP
    elif "xz compressed data" in types:
        return CompressionTypeEnum.LZMA
    elif "zstandard compressed data" in types:
        return CompressionTypeEnum.ZSTANDARD
    else:
        raise RepoManagementFileError(
            f"An error occured while attempting to retrieve the compression type of tar file: {path}!\n"
            "Unknown compression type encountered."
        )


def open_tarfile(
    path: Path,
    compression: CompressionTypeEnum | None = None,
    mode: Literal["r", "w", "x"] = "r",
) -> TarFile:
    """Open a file as a TarFile.

    This function distinguishes between bzip2, gzip, lzma and zstandard compression depending on file suffix.
    The detection can be overridden by providing either a file suffix or compression type.

    Parameters
    ----------
    path: Path
        A Path to a file
    compression: CompressionTypeEnum | None
        An optional compression type to override the detection based on mime type.
    mode: Literal["r", "w", "x"]
        A mode to open the file with (defaults to "r").
        "r" - open file for reading
        "w" - open file for writing
        "x" - create file

    Raises
    ------
    ValueError
        If the file represented by db_path does not exist
    tarfile.ReadError
        If the file could not be opened
    RepoManagementFileError
        If the compression type is unknown

    Returns
    -------
    tarfile.Tarfile
        An instance of Tarfile
    """
    debug(f"Opening file {path}...")

    if not path.is_absolute():
        raise RepoManagementFileError(f"An error occured while attempting to resolve a file path: {path} is relative!")
    if path.is_symlink():
        path = path.resolve()

    compression_type = compression if compression else compression_type_of_tarfile(path=path)

    match compression_type:
        case CompressionTypeEnum.NONE | CompressionTypeEnum.BZIP2 | CompressionTypeEnum.GZIP | CompressionTypeEnum.LZMA:
            try:
                return tarfile_open(name=path, mode=f"{mode}:{compression_type.value}")
            except ReadError as e:
                raise RepoManagementFileError(
                    f"An error occured attempting to read tar file {path} using compression type "
                    f"{compression_type.value}.\n{e}"
                )
        case CompressionTypeEnum.ZSTANDARD:
            return ZstdTarFile(name=path, mode=mode)
        case _:
            raise RepoManagementFileError(
                f"Unknown compression type {compression_type} encountered while attempting to open file {path}!"
            )


async def extract_file_from_tarfile(
    tarfile: TarFile,
    file: str,
    as_stringio: bool = False,
    gzip_compressed: bool = False,
) -> IO[bytes] | StringIO:
    """Extract a file from a TarFile and return it as a bytes stream or text I/O buffer.

    Parameters
    ----------
    tarfile: TarFile
        An instance of TarFile
    file: str
        A string representing the name of a file contained in tarfile
    as_stringio: bool
        Whether to return the data as StringIO (defaults to False)
    gzip_compressed: bool
        Whether the target file in tarfile is gzip compressed (defaults to False)

    Raises
    ------
    RepoManagementFileNotFoundError
        If the requested file does not exist in the tarfile or if the requested file is neither a file nor a symlink
    RepoManagementFileError
        If gzip_compressed is True and file is not a gzip compressed file or the gzip compressed file is corrupted

    Returns
    -------
    IO[bytes] | StringIO
        A bytes stream or StringIO that represents the file to extract
    """
    debug(f"Extracting file {file} from {str(tarfile.name)}...")

    try:
        extracted = tarfile.extractfile(file)
        if extracted is None:
            raise RepoManagementFileNotFoundError(
                f"File {file} in {str(tarfile.name)} is not a file or a symbolic link!"
            )
    except KeyError as e:
        raise RepoManagementFileNotFoundError(f"File {file} not found in {str(tarfile.name)}!\n{e}")

    if gzip_compressed:
        try:
            with gzip_open(extracted) as extracted_file:
                if as_stringio:
                    return StringIO(BytesIO(extracted_file.read()).read().decode("utf-8"))
                else:
                    return BytesIO(extracted_file.read())
        except BadGzipFile as e:
            raise RepoManagementFileError(f"An error occured trying to read the gzip compressed file {file}\n{e}\n")
    else:
        if as_stringio:
            return StringIO(initial_value=extracted.read().decode("utf-8"))
        else:
            return extracted


def names_in_tarfile(tarfile: TarFile, names: list[str] | set[str]) -> bool:
    """Check whether a list of names is found in the list of names of a TarFile.

    Parameters
    ----------
    tarfile: TarFile
        A TarFile to lookup names in
    names: list[str] | set[str]
        A list or set of names to lookup

    Returns
    -------
    bool
        True if all names can be found in the list of TarFile names, else False
    """
    if isinstance(names, list):
        names = set(names)

    if names == set([name for name in tarfile.getnames() if name in names]):
        return True
    else:
        return False


def read_text_from_file(path: str | Path) -> StringIO:
    """Read text from a file and return it in a StringIO.

    Parameters
    ----------
    path: str | Path
        A Path or str representing a file to read

    Raises
    ------
    RepoManagementFileNotFoundError
        If the file identifed by path does not exist

    Returns
    -------
    StringIO
        A string IO stream representing the contents of the file
    """
    if isinstance(path, str):
        path = Path(path)

    try:
        return StringIO(initial_value=path.read_text())
    except FileNotFoundError as e:
        raise RepoManagementFileNotFoundError(e)



"""Functions and helpers to work with files in repod managed repositories."""

from itertools import zip_longest
from logging import debug, info
from pathlib import Path
from re import Match, fullmatch
from shutil import copy2
from typing import Any

from pydantic import BaseModel, model_validator

from repod.common.enums import RepoFileEnum
from repod.common.regex import PACKAGE_FILENAME, SIGNATURE_FILENAME
from repod.errors import RepoManagementFileError


def filename_parts(file: Path) -> dict[str, str]:
    """Split a package or signature name and return its specific metadata in a dict.

    Parameters
    ----------
    file: Path
        A package filename path

    Raises
    ------
    ValueError
        If file can not be split to derive architecture and suffix information.
        If file can not be split to derive name and version information.

    Returns
    -------
    dict[str, str]
        A dict carrying data, derived from the package filename:
        - arch: CPU architecture
        - suffix: The package compression
        - version: The full version
        - pkgrel: The release version of the package
        - pkgver: The version of the package
        - epoch: The epoch of the package (if any)
        - name: The name of the package
    """
    debug(f"Retrieving data from filename {file}")
    name = file.name
    output_dict: dict[str, str] = {}

    arch_suffix_split = name.split("-")[-1].split(".")
    if not (5 >= len(arch_suffix_split) >= 3):
        raise ValueError(f"The provided file name {file} can not be split for architecture and suffix information!")

    output_dict["arch"] = arch_suffix_split[0]
    output_dict["suffix"] = ".".join(arch_suffix_split[1:])

    name_version = name.split("-")[:-1]
    if len(name_version) < 3:
        raise ValueError(f"The provided file name {file} can not be split for name and version information!")

    output_dict["version"] = "-".join([name_version[-2], name_version[-1]])
    output_dict["pkgrel"] = name_version[-1]
    output_dict["pkgver"] = name_version[-2].split(":")[-1]
    output_dict["epoch"] = name_version[-2].split(":")[0] if len(name_version[-2].split(":")) > 1 else ""
    output_dict["name"] = "-".join(name_version[0:-2])

    return output_dict


def shared_base_path(path_a: Path, path_b: Path) -> Path:
    """Return the shared base path of two absolute paths.

    Parameters
    ----------
    path_a: Path
        An absolute path
    path_b: Path
        An absolute path

    Raises
    ------
    ValueError
        If either path_a or path_b are not absolute paths

    Returns
    -------
    Path
        The shared base path of path_a and path_b
    """
    debug(f"Calculating the shared base path of {path_a} and {path_b}...")
    for path in (path_a, path_b):
        if not path.is_absolute():
            raise ValueError(f"The path {path} is not absolute!")

    output_list: list[str] = []
    # NOTE: we skip branch check here, because coveragepy does not detect that we can never have any of the Path objects
    # return an empty list of parts, as they are ensured to be absolute
    for part_a, part_b in zip_longest(path_a.parts, path_b.parts, fillvalue=None):  # pragma: no branch
        if part_a is not None and part_b is not None and part_a == part_b:
            output_list.append(part_a)
        else:
            break

    debug(f"Calculated shared base path: {Path(*output_list)}")
    return Path(*output_list)


def relative_to_shared_base(path_a: Path, path_b: Path) -> Path:
    """Return a Path to path_a, relative to the shared base path of path_a and path_b.

    This function calls shared_base_path() to determine the shared base path of path_a and path_b.

    Parameters
    ----------
    path_a: Path
        An absolute path
    path_b: Path
        An absolute path

    Returns
    -------
    Path
        The path to path_a, relative to the shared base path of path_a and path_b
    """
    debug(f"Calculating a path relative to the shared base path of {path_a} and {path_b}...")
    shared_base = shared_base_path(path_a=path_a, path_b=path_b)
    parent_distance = len(path_b.parent.parts) - len(shared_base.parts)
    debug(f"The parent distance of {path_a} to {shared_base} is {parent_distance}.")

    return_path: Path
    if parent_distance > 0:
        return_path = Path(*(["../"] * parent_distance)) / path_a.relative_to(shared_base)
    else:
        return_path = path_a.relative_to(shared_base)

    debug(f"Calculated path relative to the shared base path {shared_base}: {return_path}")
    return return_path


class RepoFile(BaseModel):
    """Class to interact with files in a repository.

    Attributes
    ----------
    file_type: RepoFileEnum
        A RepoFileEnum member defining which type of file is targeted
    file_path: Path
        The path to the file
    symlink_path: Path
        The path to a symlink to file_path
    """

    file_type: RepoFileEnum
    file_path: Path
    symlink_path: Path

    @classmethod
    def validate_path(cls, path: Path, file_type: RepoFileEnum) -> None:
        """Validate path to match regex.

        Parameters
        ----------
        path: Path
            A path to match
        file_type: RepoFileEnum
            A member of RepoFileEnum identifying which type of RepoFile to validate

        Raises
        ------
        ValueError
            If path can not be validated
        """
        match file_type:
            case RepoFileEnum.PACKAGE:
                if not path.is_absolute():
                    raise ValueError(f"The path {path} must be absolute!")
                if not isinstance(fullmatch(PACKAGE_FILENAME, path.name), Match):
                    raise ValueError(f"The path {path} does not match the regular expression {PACKAGE_FILENAME}.")
                return

            case RepoFileEnum.PACKAGE_SIGNATURE:
                if not path.is_absolute():
                    raise ValueError(f"The path {path} must be absolute!")
                if not isinstance(fullmatch(SIGNATURE_FILENAME, path.name), Match):
                    raise ValueError(f"The path {path} does not match the regular expression {SIGNATURE_FILENAME}.")
                return
            case _:
                raise RuntimeError(f"Unknown repository file type: {file_type}")

    @model_validator(mode="before")
    def validate_paths(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Validate absolute Paths.

        Parameters
        ----------
        values: dict[str, Any]
            A dict with all values of the RepoFile instance

        Raises
        ------
        ValueError
            If file_path and symlink_path are equal.
            If the path can not be validated using RepoFile.validate_path()

        Returns
        -------
        Path
            The validated Path
        """
        file_type: RepoFileEnum = values.get("file_type")  # type: ignore[assignment]

        paths: list[Path] = []
        paths.append(values.get("file_path"))  # type: ignore[arg-type]
        paths.append(values.get("symlink_path"))  # type: ignore[arg-type]

        if paths[0] == paths[1]:
            raise ValueError(
                f"The file_path ({paths[0]}) and the symlink_path ({paths[1]}) attributes must not be equal!"
            )

        for path in paths:
            RepoFile.validate_path(path=path, file_type=file_type)

        return values

    def check_file_path_exists(self, exists: bool = True) -> None:
        """Ensure that file_path exists.

        Parameters
        ----------
        exists: bool
            Whether file_path should exist or not (defaults to True)

        Raises
        ------
        RepoManagementFileError
            If self.file_path does not exist
        """
        if exists:
            if not self.file_path.exists():
                raise RepoManagementFileError(
                    f"An error occured checking for the existence of a file: {self.file_path} does not exist!"
                )
        else:
            if self.file_path.exists():
                raise RepoManagementFileError(
                    f"An error occured checking for the existence of a file: {self.file_path} exists already!"
                )

    def check_symlink_path_exists(self, exists: bool = True) -> None:
        """Ensure that symlink_path exists.

        Parameters
        ----------
        exists: bool
            Whether symlink_path should exist or not (defaults to True)

        Raises
        ------
        RepoManagementFileError
            If self.symlink_path does not exist
        """
        if exists:
            if not self.symlink_path.exists():
                raise RepoManagementFileError(
                    "An error occured checking for the existence of a symlink: {self.symlink_path} does not exist!"
                )
        else:
            if self.symlink_path.exists():
                raise RepoManagementFileError(
                    "An error occured checking for the existence of a symlink: {self.symlink_path} exists already!"
                )

    def copy_from(self, path: Path) -> None:
        """Copy file from a provided Path to file_path.

        Before doing further checks, RepoFile.validate_path() is run on path.

        Parameters
        ----------
        path: Path
            Path to move from

        Raises
        ------
        RepoManagementFileError
            If path does not exist
        """
        info(f"Copy {self.file_path} from {path}...")
        RepoFile.validate_path(path=path, file_type=self.file_type)
        if not path.exists():
            raise RepoManagementFileError(f"Error on trying to move file: The input file {path} does not exist!")

        self.check_file_path_exists(exists=False)
        copy2(src=path, dst=self.file_path)

    def move_from(self, path: Path) -> None:
        """Move file from a provided Path to file_path.

        Before doing further checks, RepoFile.validate_path() is run on path.

        Parameters
        ----------
        path: Path
            Path to move from

        Raises
        ------
        RepoManagementFileError
            If path does not exist
        """
        info(f"Move {self.file_path} from {path}...")
        RepoFile.validate_path(path=path, file_type=self.file_type)
        if not path.exists():
            raise RepoManagementFileError(f"Error on trying to move file: The input file {path} does not exist!")

        self.check_file_path_exists(exists=False)
        path.rename(target=self.file_path)

    def link(self, check: bool = True) -> None:
        """Link the symlink_path to file_path using a relative symlink.

        Parameters
        ----------
        check: bool
            Whether to check if the symlink_path exists already prior to linking (defaults to True)

        Raises
        ------
        RepoManagementFileError
            If the file exists already
        """
        info(f"Link {self.symlink_path} to {self.file_path}...")
        if check:
            debug(f"Checking that {self.symlink_path} does not yet exist...")
            self.check_symlink_path_exists(exists=False)

        try:
            self.symlink_path.symlink_to(relative_to_shared_base(path_a=self.file_path, path_b=self.symlink_path))
        except FileExistsError:
            raise RepoManagementFileError(
                f"An error occured attempting to symlink {self.symlink_path} to {self.file_path}."
                f"{self.symlink_path} exists already!"
            )

    def unlink(self, check: bool = True) -> None:
        """Unlink the symlink_path from file_path.

        Parameters
        ----------
        check: bool
            Whether to check that the symlink_path does not exist prior to removal (defaults to True)

        Raises
        ------
        RepoManagementFileError
            If the file does not exist and check = True
        """
        info(f"Unlink {self.symlink_path} from {self.file_path}...")
        if check:
            debug(f"Checking that {self.symlink_path} exists...")
            self.check_symlink_path_exists()

        self.symlink_path.unlink(missing_ok=not check)

    def remove(self, force: bool = False, unlink: bool = False) -> None:
        """Remove file_path and optionally unlink symlink_path from file_path.

        Parameters
        ----------
        force: bool
            Whether to not check for path existence before unlinking and to ignore errors on removing non-existing files
            (defaults to False)
        """
        info(f"Removing {self.file_path}...")
        if not force:
            debug(f"Checking that {self.file_path} exists...")
            self.check_file_path_exists()

        self.file_path.unlink(missing_ok=force)

        if unlink:
            info(f"Removing {self.symlink_path}...")
            if not force:
                debug(f"Checking that {self.symlink_path} exists...")
                self.check_symlink_path_exists()
            self.unlink(check=not force)



"""Sync database handling."""

from __future__ import annotations

import io
import re
from enum import IntEnum
from logging import debug, warning
from pathlib import Path
from tarfile import DIRTYPE, TarFile, TarInfo
from time import time

from jinja2 import Environment, PackageLoader, TemplateNotFound
from orjson import OPT_APPEND_NEWLINE, OPT_INDENT_2, OPT_SORT_KEYS, dumps
from pydantic import BaseModel, ValidationError

from repod.common.enums import (
    CompressionTypeEnum,
    FieldTypeEnum,
    FilesVersionEnum,
    PackageDescVersionEnum,
)
from repod.common.models import (
    Arch,
    Backup,
    Base,
    BuildDate,
    CheckDepends,
    Conflicts,
    CSize,
    Depends,
    Desc,
    FileList,
    FileName,
    Groups,
    ISize,
    License,
    MakeDepends,
    Md5Sum,
    Name,
    OptDepends,
    Packager,
    PgpSig,
    Provides,
    Replaces,
    SchemaVersionV1,
    SchemaVersionV2,
    Sha256Sum,
    Url,
    Version,
)
from repod.errors import (
    RepoManagementFileError,
    RepoManagementFileNotFoundError,
    RepoManagementValidationError,
)
from repod.files.common import open_tarfile
from repod.repo.management import outputpackage

DB_USER = "root"
DB_GROUP = "root"
DB_FILE_MODE = "0644"
DB_DIR_MODE = "0755"
DESC_JSON: dict[str, tuple[str, FieldTypeEnum]] = {
    "%BASE%": ("base", FieldTypeEnum.STRING),
    "%VERSION%": ("version", FieldTypeEnum.STRING),
    "%MAKEDEPENDS%": ("makedepends", FieldTypeEnum.STRING_LIST),
    "%CHECKDEPENDS%": ("checkdepends", FieldTypeEnum.STRING_LIST),
    "%FILENAME%": ("filename", FieldTypeEnum.STRING),
    "%NAME%": ("name", FieldTypeEnum.STRING),
    "%DESC%": ("desc", FieldTypeEnum.STRING),
    "%GROUPS%": ("groups", FieldTypeEnum.STRING_LIST),
    "%CSIZE%": ("csize", FieldTypeEnum.INT),
    "%ISIZE%": ("isize", FieldTypeEnum.INT),
    "%MD5SUM%": ("md5sum", FieldTypeEnum.STRING),
    "%SHA256SUM%": ("sha256sum", FieldTypeEnum.STRING),
    "%PGPSIG%": ("pgpsig", FieldTypeEnum.STRING),
    "%URL%": ("url", FieldTypeEnum.STRING),
    "%LICENSE%": ("license", FieldTypeEnum.STRING_LIST),
    "%ARCH%": ("arch", FieldTypeEnum.STRING),
    "%BUILDDATE%": ("builddate", FieldTypeEnum.INT),
    "%PACKAGER%": ("packager", FieldTypeEnum.STRING),
    "%REPLACES%": ("replaces", FieldTypeEnum.STRING_LIST),
    "%CONFLICTS%": ("conflicts", FieldTypeEnum.STRING_LIST),
    "%PROVIDES%": ("provides", FieldTypeEnum.STRING_LIST),
    "%DEPENDS%": ("depends", FieldTypeEnum.STRING_LIST),
    "%OPTDEPENDS%": ("optdepends", FieldTypeEnum.STRING_LIST),
    "%BACKUP%": ("backup", FieldTypeEnum.STRING_LIST),
}
FILES_JSON: dict[str, tuple[str, FieldTypeEnum]] = {"%FILES%": ("files", FieldTypeEnum.STRING_LIST)}
FILES_VERSIONS: dict[int, dict[str, set[str]]] = {
    1: {
        "required": {
            "files",
        },
        "optional": set(),
    },
}
DEFAULT_FILES_VERSION = 1
DEFAULT_PACKAGE_DESC_VERSION = 2
PACKAGE_DESC_VERSIONS: dict[int, dict[str, set[str] | int]] = {
    1: {
        "required": {
            "arch",
            "base",
            "builddate",
            "csize",
            "desc",
            "filename",
            "isize",
            "license",
            "md5sum",
            "name",
            "packager",
            "sha256sum",
            "url",
            "version",
        },
        "optional": {
            "checkdepends",
            "conflicts",
            "depends",
            "backup",
            "groups",
            "makedepends",
            "optdepends",
            "pgpsig",
            "provides",
            "replaces",
        },
        "output_package_version": 1,
        "output_package_base_version": 1,
    },
    2: {
        "required": {
            "arch",
            "base",
            "builddate",
            "csize",
            "desc",
            "filename",
            "isize",
            "license",
            "name",
            "packager",
            "sha256sum",
            "url",
            "version",
        },
        "optional": {
            "checkdepends",
            "conflicts",
            "depends",
            "backup",
            "groups",
            "makedepends",
            "optdepends",
            "pgpsig",
            "provides",
            "replaces",
        },
        "output_package_version": 2,
        "output_package_base_version": 1,
    },
}


class RepoDbMemberTypeEnum(IntEnum):
    """An IntEnum to distinguish different types of files in a repository sync database.

    Attributes
    ----------
    UNKNOWN: int
        An unsupported file of unknown type
    DESC: int
        A 'desc' description file of a package
    FILES: int
        A 'files' file of a package
    """

    UNKNOWN = 0
    DESC = 1
    FILES = 2


class RepoDbTypeEnum(IntEnum):
    """An IntEnum to distinguish types of binary repository database files.

    Attributes
    ----------
    DEFAULT: int
        Use this to identify .db files
    FILES: int
        Use this to identify .files files
    """

    DEFAULT = 0
    FILES = 2


def get_desc_json_keys() -> set[str]:
    """Get the keys of repod.models.repo.DESC_JSON.

    Returns
    -------
    set[str]
        A set of strings representing the keys of repod.models.repo.DESC_JSON
    """
    return set(DESC_JSON.keys())


def get_desc_json_name(key: str) -> str:
    """Get the JSON name of a given key from the definition in repod.models.repo.DESC_JSON.

    Parameters
    ----------
    key: str
        A string corresponding with one of the keys in repod.models.repo.DESC_JSON

    Raises
    ------
    RepoManagementFileError
        If an unknown key is encountered

    Returns
    -------
    str
        The JSON name of a given 'desc' file identifier provided by key
    """
    try:
        return DESC_JSON[key][0]
    except KeyError:
        raise RepoManagementFileError(f"The key {key} is not a known 'desc' file identifier.")


def get_desc_json_field_type(key: str) -> FieldTypeEnum:
    """Get the FieldTypeEnum of a given key from the definition in repod.models.repo.DESC_JSON.

    Parameters
    ----------
    key: str
        A string corresponding with one of the keys in repod.models.repo.DESC_JSON

    Raises
    ------
    RepoManagementFileError
        If an unknown key is encountered

    Returns
    -------
    str
        The FieldTypeEnum of a given 'desc' file identifier provided by key
    """
    try:
        return DESC_JSON[key][1]
    except KeyError:
        raise RepoManagementFileError(f"The key {key} is not a known 'desc' file identifier.")


def get_files_json_keys() -> set[str]:
    """Get the keys of repod.models.repo.FILES_JSON.

    Returns
    -------
    set[str]
        A set of strings representing the keys of repod.models.repo.FILES_JSON
    """
    return set(FILES_JSON.keys())


def get_files_json_name(key: str) -> str:
    """Get the JSON name of a given key from the definition in repod.models.repo.FILES_JSON.

    Parameters
    ----------
    key: str
        A string corresponding with one of the keys in repod.models.repo.FILES_JSON

    Raises
    ------
    RepoManagementFileError
        If an unknown key is encountered

    Returns
    -------
    str
        The JSON name of a given 'files' file identifier provided by key
    """
    try:
        return FILES_JSON[key][0]
    except KeyError:
        raise RepoManagementFileError(f"The key {key} is not a known 'files' file identifier.")


def get_files_json_field_type(key: str) -> FieldTypeEnum:
    """Get the FieldTypeEnum of a given key from the definition in repod.models.repo.FILES_JSON.

    Parameters
    ----------
    key: str
        A string corresponding with one of the keys in repod.models.repo.FILES_JSON

    Raises
    ------
    RepoManagementFileError
        If an unknown key is encountered

    Returns
    -------
    str
        The FieldTypeEnum of a given 'desc' file identifier provided by key
    """
    try:
        return FILES_JSON[key][1]
    except KeyError:
        raise RepoManagementFileError(f"The key {key} is not a known 'files' file identifier.")


class SyncDatabase(BaseModel):
    """A model describing a repository sync database.

    Attributes
    ----------
    database: Path
        The path to the database that the instance manages
    database_type: RepoDbTypeEnum
        The type of database that the instance manages
    compression_type: CompressionTypeEnum
        The compression type which is used for  the database
    """

    database: Path
    database_type: RepoDbTypeEnum | None = None
    compression_type: CompressionTypeEnum | None = None
    desc_version: PackageDescVersionEnum
    files_version: FilesVersionEnum

    @classmethod
    async def outputpackagebase_to_tarfile(
        cls,
        tarfile: TarFile,
        database_type: RepoDbTypeEnum | None,
        model: outputpackage.OutputPackageBase,
        packagedesc_version: PackageDescVersionEnum,
        files_version: FilesVersionEnum,
    ) -> None:
        """Stream descriptor files derived from an OutputPackageBase to a TarFile.

        Allows streaming to a default repository database or a files database

        Parameters
        ----------
        tarfile: TarFile
            A TarFile to stream data to
        db_type: RepoDbTypeEnum
            The type of database to stream to
        model: OutputPackageBase
            The OutputPackageBase instance to derive descriptor files from
        packagedesc_version: PackageDescVersionEnum
            The version of PackageDesc to use
        files_version: FilesVersionEnum
            The version of Files to use
        """
        for desc_model, files_model in await model.get_packages_as_models(
            packagedesc_version=packagedesc_version,
            files_version=files_version,
        ):
            dirname = f"{desc_model.get_name()}-{model.get_version()}"
            directory = TarInfo(dirname)
            directory.type = DIRTYPE
            directory.mtime = int(time())
            directory.uname = DB_USER
            directory.gname = DB_GROUP
            directory.mode = int(DB_DIR_MODE, base=8)
            tarfile.addfile(directory)

            desc_content = io.StringIO()
            await desc_model.render(output=desc_content)
            desc_file = TarInfo(f"{dirname}/desc")
            desc_file.size = len(desc_content.getvalue().encode())
            desc_file.mtime = int(time())
            desc_file.uname = DB_USER
            desc_file.gname = DB_GROUP
            desc_file.mode = int(DB_FILE_MODE, base=8)
            tarfile.addfile(desc_file, io.BytesIO(desc_content.getvalue().encode()))

            if database_type == RepoDbTypeEnum.FILES:
                files_content = io.StringIO()
                await files_model.render(output=files_content)
                files_file = TarInfo(f"{dirname}/files")
                files_file.size = len(files_content.getvalue().encode())
                files_file.mtime = int(time())
                files_file.uname = DB_USER
                files_file.gname = DB_GROUP
                files_file.mode = int(DB_FILE_MODE, base=8)
                tarfile.addfile(files_file, io.BytesIO(files_content.getvalue().encode()))

    async def add(self, model: outputpackage.OutputPackageBase) -> None:
        """Write descriptor files for packages of a single pkgbase to the repository sync database.

        Parameters
        ----------
        model: OutputPackageBase
            The model to use for streaming descriptor files to the repository database
        """
        debug(f"Opening file {self.database} for writing...")

        with open_tarfile(
            path=self.database,
            compression=self.compression_type,
            mode="w",
        ) as database_file:
            await SyncDatabase.outputpackagebase_to_tarfile(
                tarfile=database_file,
                database_type=self.database_type,
                model=model,
                packagedesc_version=self.desc_version,
                files_version=self.files_version,
            )

    async def outputpackagebases(self) -> list[tuple[str, outputpackage.OutputPackageBase]]:
        """Read a repo sync database and return the name of each pkgbase and respective data.

        Returns
        -------
        list[tuple[str, OutputPackageBase]]:
            A tuple holding the name of a pkgbase and its accompanying data in an instance of OutputPackageBase
        """
        packages: dict[str, outputpackage.OutputPackageBase] = {}
        package_descs: dict[str, PackageDesc] = {}
        package_files: dict[str, Files] = {}

        with open_tarfile(path=self.database, compression=self.compression_type) as db_file:
            for file_name in [
                file_name for file_name in db_file.getnames() if re.search(r"(/desc|/files)$", file_name)
            ]:
                pkgname = "".join(re.split("(-)", re.sub(r"(/desc|/files)$", "", file_name))[:-4])

                if re.search("(/desc)$", file_name):
                    package_descs.update(
                        {
                            pkgname: await PackageDesc.from_stream(
                                data=io.StringIO(
                                    io.BytesIO(
                                        db_file.extractfile(file_name).read(),  # type: ignore[union-attr]
                                    )
                                    .read()
                                    .decode("utf-8"),
                                )
                            ),
                        },
                    )
                elif re.search("(/files)$", file_name):
                    package_files.update(
                        {
                            pkgname: await Files.from_stream(
                                data=io.StringIO(
                                    io.BytesIO(
                                        db_file.extractfile(file_name).read(),  # type: ignore[union-attr]
                                    )
                                    .read()
                                    .decode("utf-8"),
                                )
                            )
                        },
                    )
                else:  # pragma: no cover
                    # NOTE: this branch can never be reached, but we add it to make tests happy
                    raise RuntimeError(
                        f"The database file {self.database} contains the member {file_name} of an unsupported type!"
                    )

        for name, package_desc in package_descs.items():
            if packages.get(package_desc.get_base()):
                packages[package_desc.get_base()].add_packages(
                    [package_desc.get_output_package(files=package_files.get(name))]
                )
            else:
                packages.update(
                    {
                        package_desc.get_base(): package_desc.get_output_package_base(files=package_files.get(name)),
                    }
                )

        return list(packages.items())

    async def stream_management_repo(self, path: Path) -> None:
        """Stream descriptor files read from JSON files of a management repository to the repository sync database.

        Parameters
        ----------
        path: Path
            The directory containing the files of the management repository
        """
        file_list = sorted(path.glob("*.json"))
        if not file_list:
            debug(f"There are no JSON files in {path}! Creating empty sync db.")

        with open_tarfile(self.database, compression=self.compression_type, mode="w") as database_file:
            for json_file in file_list:
                await SyncDatabase.outputpackagebase_to_tarfile(
                    tarfile=database_file,
                    database_type=self.database_type,
                    model=await outputpackage.OutputPackageBase.from_file(path=json_file),
                    packagedesc_version=self.desc_version,
                    files_version=self.files_version,
                )


class PackageDesc(BaseModel):
    """A template class with helper methods to create instances of one of its (versioned) subclasses.

    This class should not be instantiated directly, as it only provides generic instance methods for its subclasses.

    NOTE: The `from_dict()` classmethod is used to create one of the versioned subclasses.
    """

    @classmethod
    def from_dict(cls, data: dict[str, int | str | list[str]]) -> PackageDesc:
        """Create an instance of one of PackageDesc's subclasses from a dict.

        This method should be used with data derived from reading a 'desc' file from a repository sync database.

        Parameters
        ----------
        data: dict[str, int | str | list[str]]
            A dict containing data required to instantiate a subclass of PackageDesc

        Raises
        ------
        RepoManagementValidationError
            If a ValidationError is encountered while instantiating one of PackageDesc's subclasses
            If no supported schema_version can be found

        Returns
        -------
        PackageDesc
            An instance of one of the subclasses of PackageDesc
        """

        def derive_package_desc_version(data: set[str]) -> int | None:
            """Derive which PackageDesc subclass to instantiate.

            Parameters
            ----------
            data: set[str]
                 A set of strings representing the keys of the data dict passed to PackageDesc.from_dict()

            Returns
            -------
            int | None
                The integer representing the version of the PackageDesc subclass, else None
            """
            for version, config in sorted(PACKAGE_DESC_VERSIONS.items(), reverse=True):
                debug(f"Comparing 'desc' data to schema version {version}")
                config_required: set[str] = config["required"]  # type: ignore[assignment]
                config_optional: set[str] = config["optional"]  # type: ignore[assignment]
                if config_required.issubset(data):
                    optionals = data - config_required

                    if len(optionals) > len(config_optional):
                        continue
                    else:
                        unmatched_optional = 0
                        for optional in optionals:
                            if optional not in config_optional:
                                unmatched_optional += 1

                    if unmatched_optional != 0:
                        continue

                    return version
                else:
                    continue

            return None

        debug(f"Creating Files from data: {data}")
        detected_version = derive_package_desc_version(data=set(data.keys()))
        debug(f"Detected schema version of 'desc' file: {detected_version}")

        if detected_version is not None and (detected_version < DEFAULT_PACKAGE_DESC_VERSION):
            warning(f"Detected 'desc' version {detected_version}, but {DEFAULT_PACKAGE_DESC_VERSION} is the default!")

        match detected_version:
            case 1:
                try:
                    return PackageDescV1.model_validate(data)
                except ValidationError as e:
                    raise RepoManagementValidationError(
                        "A validation error occured while attempting to instantiate a PackageDescV1 using the data:\n"
                        f"{data}\n{e}"
                    )
            case 2:
                try:
                    return PackageDescV2.model_validate(data)
                except ValidationError as e:
                    raise RepoManagementValidationError(
                        "A validation error occured while attempting to instantiate a PackageDescV2 using the data:\n"
                        f"{data}\n{e}"
                    )
            case _:
                raise RepoManagementValidationError(f"The data format of the 'desc' file is unknown:\n{data}")

    @classmethod
    async def from_stream(cls, data: io.StringIO) -> PackageDesc:
        """Initialize a PackageDesc from an input stream.

        Parameters
        ----------
        data: io.StringIO
            A buffered I/O that represents a 'desc' file

        Raises
        ------
        errors.RepoManagementValidationError
            If a pydantic.error_wrappers.ValidationError is raised (e.g. due to a missing attribute) or if a ValueError
            is raised when converting data

        Returns
        -------
        PackageDesc
            A PackageDesc instance based on data
        """
        current_header = ""
        current_type: FieldTypeEnum
        int_types: dict[str, int] = {}
        string_types: dict[str, str] = {}
        string_list_types: dict[str, list[str]] = {}
        keys = get_desc_json_keys()

        for line in data:
            line = line.strip()
            if not line:
                continue

            if line in keys:
                current_header = get_desc_json_name(key=line)
                current_type = get_desc_json_field_type(line)
                # FIXME: find better way to provide a default
                # (None or empty list for STRING_LIST as they all are list[str] | None)
                if current_header and current_type == FieldTypeEnum.STRING_LIST:
                    string_list_types[current_header] = []

                continue

            if current_header:
                try:
                    match current_type:
                        case FieldTypeEnum.STRING_LIST:
                            string_list_types[current_header] += [line]
                        case FieldTypeEnum.STRING:
                            string_types[current_header] = line
                        case FieldTypeEnum.INT:
                            int_types[current_header] = int(line)
                        case _:  # pragma: no cover
                            pass
                except ValueError as e:
                    raise RepoManagementValidationError(
                        "An error occured while attempting to cast values for a 'desc' file!\n"
                        f"\n{data.getvalue()}\n{e}"
                    )

        desc_dict: dict[str, int | str | list[str]] = {**int_types, **string_types, **string_list_types}
        try:
            return PackageDesc.from_dict(desc_dict)
        except RepoManagementValidationError as e:
            raise RepoManagementValidationError(
                "An error occured while validating a 'desc' file!\n" f"\n{data.getvalue()}\n{e}"
            )

    async def render(self, output: io.StringIO) -> None:
        """Use a 'desc' jinja template to write the PackageDesc contents to an output stream.

        Parameters
        ----------
        output: io.StringIO
            An output stream to write to

        Raises
        ------
        RepoManagementFileNotFoundError
            If no matching template can be found
        """
        # NOTE: We are not rendering HTML and need special characters, hence we are not affected by XSS problems and set
        # autoescape=False
        env = Environment(  # nosec: B701
            autoescape=False,
            loader=PackageLoader("repod", "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
            enable_async=True,
        )
        template_file = f"desc_v{self.get_schema_version()}.j2"

        debug(f"Rendering PackageDesc data using template file {template_file}...")

        try:
            template = env.get_template(template_file)
        except TemplateNotFound:
            raise RepoManagementFileNotFoundError(f"The 'desc' template file {template_file} could not be found!")
        output.write(await template.render_async(self.model_dump(mode="json")))

    def get_output_package(self, files: Files | None) -> outputpackage.OutputPackage:
        """Transform the PackageDesc model and an optional Files model into an OutputPackage model.

        NOTE: This method only successfully returns if the instance of the class using it defines the required fields!
        The PackageDesc class does not do that!

        Parameters
        ----------
        files: Files | None:
            A pydantic model, that represents the list of files, that belong to the package described by self

        Raises
        ------
        RuntimeError
            If called on the PackageDesc template class
            If there is no matching OutputPackage schema_version defined in the configuration for the given
            PackageDesc.schema_version (bug in PACKAGE_DESC_VERSIONS).
        RepoManagementValidationError
            If no configuration is available for a given PackageDesc.schema_version

        Returns
        -------
        OutputPackage
            A pydantic model, that describes a package and its list of files
        """
        # TODO: write detection version for
        schema_version = self.get_schema_version()
        schema_config = PACKAGE_DESC_VERSIONS.get(schema_version)
        desc_dict = self.model_dump(mode="json")
        if schema_config:
            output_package_version = schema_config.get("output_package_version")
            match output_package_version:
                case 1:
                    if files:
                        desc_dict["files"] = files

                    return outputpackage.OutputPackageV1.model_validate(desc_dict)
                case 2:
                    if files:
                        desc_dict["files"] = files

                    return outputpackage.OutputPackageV2.model_validate(desc_dict)
                case _:
                    raise RuntimeError(
                        f"The OutputPackage version '{output_package_version}' is not valid. This is a bug!"
                    )
        else:
            raise RepoManagementValidationError(
                f"The schema version '{schema_version}' is not valid for creating a "
                "OutputPackage from a PackageDesc!"
            )

    def get_output_package_base(self, files: Files | None) -> outputpackage.OutputPackageBase:
        """Transform the PackageDesc model and an Files model into an OutputPackageBase model.

        NOTE: This method only successfully returns if the instance of the class using it defines the required fields!
        The PackageDesc class does not do that!

        Parameters
        ----------
        files: Files | None:
            A pydantic model, that represents the list of files, that belong to the package described by self

        Raises
        ------
        RuntimeError
            If called on the PackageDesc template class
            If there is no matching OutputPackage schema_version defined in the configuration for the given
            PackageDesc.schema_version (bug in PACKAGE_DESC_VERSIONS).
        RepoManagementValidationError
            If no configuration is available for a given PackageDesc.schema_version

        Returns
        -------
        OutputPackageBase
            A pydantic model, that describes a package base and one of it's split packages
        """
        schema_version = self.get_schema_version()
        schema_config = PACKAGE_DESC_VERSIONS.get(schema_version)
        desc_dict = self.model_dump(mode="json")
        del desc_dict["schema_version"]
        if schema_config:
            output_package_version = schema_config.get("output_package_base_version")
            match output_package_version:
                case 1:
                    desc_dict.update({"packages": [self.get_output_package(files=files)]})
                    return outputpackage.OutputPackageBaseV1.model_validate(desc_dict)
                case _:
                    raise RuntimeError(
                        f"The OutputPackageBase version '{output_package_version}' is not valid. This is a bug!"
                    )
        else:
            raise RepoManagementValidationError(
                f"The schema version '{schema_version}' is not valid for creating a "
                "OutputPackageBase from a PackageDesc!"
            )

    def get_base(self) -> str:
        """Get the base attribute of the PackageDesc instance.

        NOTE: This method only successfully returns if the instance of the class using it defines the `base` field! The
        PackageDesc class does not do that!

        Raises
        ------
        RuntimeError
            If the method is called on an instance of the PackageDesc template class

        Returns
        -------
        str
            The base attribute of a PackageDesc subclass
        """
        if hasattr(self, "base"):
            return str(self.base)
        else:
            raise RuntimeError(
                "It is not possible to retrieve the 'base' attribute from the template class PackageDesc!"
            )

    def get_name(self) -> str:
        """Get the name attribute of the PackageDesc instance.

        NOTE: This method only successfully returns if the instance of the class using it defines the `name` field! The
        PackageDesc class does not do that!

        Raises
        ------
        RuntimeError
            If the method is called on an instance of the PackageDesc template class

        Returns
        -------
        str
            The name attribute of a PackageDesc subclass
        """
        if hasattr(self, "name"):
            return str(self.name)
        else:
            raise RuntimeError(
                "It is not possible to retrieve the 'name' attribute from the template class PackageDesc!"
            )

    def get_schema_version(self) -> int:
        """Get the schema_version attribute of the PackageDesc instance.

        NOTE: This method only successfully returns if the instance of the class using it defines the `schema_version`
        field! The PackageDesc class does not do that!

        Raises
        ------
        RuntimeError
            If the method is called on an instance of the PackageDesc template class

        Returns
        -------
        int
            The schema_version attribute of a PackageDesc subclass
        """
        if hasattr(self, "schema_version"):
            return int(self.schema_version)
        else:
            raise RuntimeError(
                "It is not possible to retrieve the 'schema_version' attribute from the templatet class PackageDesc!"
            )


class Files(BaseModel):
    """A template class to describe files in the context of 'files' files in a repository sync database.

    This class should not be instantiated directly, as it only provides generic instance methods for its subclasses.

    NOTE: The `from_dict()` classmethod is used to create one of the versioned subclasses.
    """

    @classmethod
    def from_dict(cls, data: dict[str, list[str]]) -> Files:
        """Class method to create one of Files' subclasses from a dict.

        This method is supposed to be called with data derived from a 'files' file.

        Parameters
        ----------
        data: dict[str, list[str]]
            A dict with required data read from a 'files' file, which is used to instantiate one of Files' subclasses

        Raises
        ------
        RepoManagementValidationError
            If a ValidationError occurs when instatiating one of Files' subclasses or if no valid schema version for the
            data could be derived.

        Returns
        -------
        Files
            A Files instance (technically only one of its subclasses), instantiated from data
        """

        def derive_files_version(data: set[str]) -> int | None:
            """Derive which Files subclass to instantiate.

            Parameters
            ----------
            data: set[str]
                 A set of strings representing the keys of the data dict passed to Files.from_dict()

            Returns
            -------
            int | None
                The integer representing the schema version of the Files subclass, else None
            """
            for version, config in sorted(FILES_VERSIONS.items(), reverse=True):
                debug(f"Comparing 'files' data to schema version {version}")
                config_required = config["required"]
                config_optional = config["optional"]
                if config_required.issubset(data):
                    optionals = data - config_required

                    if len(optionals) > len(config_optional):
                        continue
                    else:
                        unmatched_optional = 0
                        for optional in optionals:
                            if optional not in config_optional:
                                unmatched_optional += 1

                    if unmatched_optional != 0:
                        continue

                    return version
                else:
                    continue

            return None

        detected_version = derive_files_version(data=set(data.keys()))
        debug(f"Detected schema version of 'files' file: {detected_version}")

        if detected_version is not None and (detected_version < DEFAULT_FILES_VERSION):
            warning(f"Detected 'files' version {detected_version}, but {DEFAULT_FILES_VERSION} is the default!")

        match detected_version:
            case 1:
                try:
                    return FilesV1.model_validate(data)
                except ValidationError as e:
                    raise RepoManagementValidationError(
                        "A validation error occured while attempting to instantiate a FilesV1 using the data:\n"
                        f"{data}\n{e}"
                    )
            case _:
                raise RepoManagementValidationError(
                    f"The data format '{detected_version}' of the 'files' file is unknown:\n{data}"
                )

    @classmethod
    async def from_stream(cls, data: io.StringIO) -> Files:
        """Initialize a Files from an input stream.

        Parameters
        ----------
        data: io.StringIO
            A buffered I/O that represents the contents of a 'files' file

        Raises
        ------
        errors.RepoManagementValidationError
            If a pydantic.error_wrappers.ValidationError is raised (e.g. due to a missing attribute)

        Returns
        -------
        Files
            A Files instance based on data
        """
        current_header = ""
        current_type: FieldTypeEnum
        string_list_types: dict[str, list[str]] = {}
        keys = get_files_json_keys()

        for line in data:
            line = line.strip()
            if not line:
                continue

            if line in keys:
                current_header = get_files_json_name(key=line)
                current_type = get_files_json_field_type(line)
                debug(f"Found header {current_header} of type {current_type}")
                string_list_types[current_header] = []
                continue

            if current_header:
                debug(f"Adding value {line} for header {current_header}")
                string_list_types[current_header] += [line]

        try:
            return Files.from_dict(data=string_list_types)
        except RepoManagementValidationError as e:
            raise RepoManagementValidationError(
                f"An error occured while validating a 'files' file!\n" f"\n{data.getvalue()}\n{e}"
            )

    async def render(self, output: io.StringIO) -> None:
        """Use a 'files' jinja template to write the Files contents to an output stream.

        Parameters
        ----------
        output: io.StringIO
            An output stream to write to

        Raises
        ------
        RepoManagementFileNotFoundError
            If no matching template can be found
        """
        # NOTE: We are not rendering HTML and need special characters, hence we are not affected by XSS problems and set
        # autoescape=False
        env = Environment(  # nosec: B701
            loader=PackageLoader("repod", "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
            enable_async=True,
        )
        template_file = f"files_v{self.get_schema_version()}.j2"

        debug(f"Rendering Files data using template file {template_file}...")

        try:
            template = env.get_template(template_file)
        except TemplateNotFound:
            raise RepoManagementFileNotFoundError(f"The 'files' template file {template_file} could not be found!")
        output.write(await template.render_async(self.model_dump(mode="json")))

    def get_schema_version(self) -> int:
        """Get the schema_version of the Files instance.

        NOTE: This method only successfully returns if the instance of the class using it actually defines the
        `schema_version` field!

        Raises
        ------
        RuntimeError
            If the method is called on an instance of the Files template class

        Returns
        -------
        int
            The schema_version attribute of a Files subclass
        """
        if hasattr(self, "schema_version"):
            return int(self.schema_version)
        else:
            raise RuntimeError(
                "It is not possible to retrieve the 'schema_version' attribute from the Files template class!"
            )


class FilesV1(Files, FileList, SchemaVersionV1):
    """A pydantic model to describe files in the context of 'files' files in a repository sync database (version 1).

    Attributes
    ----------
    files: list[str] | None
        An optional list of files. This is the data below a %FILES% identifier in a 'files' file, which identifies which
        file(s) belong to a package
    """

    pass


class PackageDescV1(
    PackageDesc,
    Arch,
    Backup,
    Base,
    BuildDate,
    CheckDepends,
    Conflicts,
    CSize,
    Depends,
    Desc,
    FileName,
    Groups,
    ISize,
    License,
    MakeDepends,
    Md5Sum,
    Name,
    OptDepends,
    Packager,
    PgpSig,
    Provides,
    Replaces,
    SchemaVersionV1,
    Sha256Sum,
    Url,
    Version,
):
    """A model describing all identifiers in a 'desc' file (version 1).

    Attributes
    ----------
    arch: str
        The attribute can be used to describe the (required) data below an %ARCH% identifier in a 'desc' file, which
        identifies a package's architecture
    backup: list[str] | None
        The attribute can be used to describe the (optional) data below a %BACKUP% identifier in a 'desc' file, which
        identifies which file(s) of a package pacman will create backups for
    base: str
        The attribute can be used to describe the (required) data below a %BASE% identifier in a 'desc' file, which
        identifies a package's pkgbase
    builddate: int
        The attribute can be used to describe the (required) data below a %BUILDDATE% identifier in a 'desc' file,
        which identifies a package's build date (represented in seconds since the epoch)
    checkdepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %CHECKDEPENDS% identifier in a 'desc' file,
        which identifies a package's checkdepends
    conflicts: list[str] | None
        The attribute can be used to describe the (optional) data below a %CONFLICTS% identifier in a 'desc' file, which
        identifies what other package(s) a package conflicts with
    csize: int
        The attribute can be used to describe the (required) data below a %CSIZE% identifier in a 'desc' file, which
        identifies a package's size
    depends: list[str] | None
        The attribute can be used to describe the (optional) data below a %DEPENDS% identifier in a 'desc' file, which
        identifies what other package(s) a package depends on
    desc: str
        The attribute can be used to describe the (required) data below a %DESC% identifier in a 'desc' file, which
        identifies a package's description
    filename: str
        The attribute can be used to describe the (required) data below a %FILENAME% identifier in a 'desc' file, which
        identifies a package's file name
    groups: list[str] | None
        The attribute can be used to describe the (optional) data below a %GROUPS% identifier in a 'desc' file, which
        identifies a package's groups
    isize: int
        The attribute can be used to describe the (required) data below an %ISIZE% identifier in a 'desc' file, which
        identifies a package's installed size
    license: list[str]
        The attribute can be used to describe the (required) data below a %LICENSE% identifier in a 'desc' file, which
        identifies a package's license(s)
    makedepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %MAKEDEPENDS% identifier in a 'desc' file,
        which identifies a package's makedepends
    md5sum: str
        The attribute can be used to describe the (required) data below an %MD5SUM% identifier in a 'desc' file, which
        identifies a package's md5 checksum
    name: str
        The attribute can be used to describe the (required) data below a %NAME% identifier in a 'desc' file, which
        identifies a package's name
    optdepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %OPTDEPENDS% identifier in a 'desc' file,
        which identifies what other package(s) a package optionally depends on
    packager: str
        The attribute can be used to describe the (required) data below a %PACKAGER% identifier in a 'desc' file, which
        identifies a package's packager
    pgpsig: str
        The attribute can be used to describe the data below a %PGPSIG% identifier in a 'desc' file, which identifies a
        package's PGP signature
    provides: list[str] | None
        The attribute can be used to describe the (optional) data below a %PROVIDES% identifier in a 'desc' file, which
        identifies what other package(s) a package provides
    replaces: list[str] | None
        The attribute can be used to describe the (optional) data below a %REPLACES% identifier in a 'desc' file, which
        identifies what other package(s) a package replaces
    schema_version: PositiveInt
        A positive integer - 1 - identifying the schema version of the object
    sha256sum: str
        The attribute can be used to describe the (required) data below an %SHA256SUM% identifier in a 'desc' file,
        which identifies a package's sha256 checksum
    url: str
        The attribute can be used to describe the (required) data below a %URL% identifier in a 'desc' file, which
        identifies a package's URL
    version: str
        The attribute can be used to describe the (required) data below a %VERSION% identifier in a 'desc' file, which
        identifies a package's version (this is the accumulation of epoch, pkgver and pkgrel)
    """


class PackageDescV2(
    PackageDesc,
    Arch,
    Backup,
    Base,
    BuildDate,
    CheckDepends,
    Conflicts,
    CSize,
    Depends,
    Desc,
    FileName,
    Groups,
    ISize,
    License,
    MakeDepends,
    Name,
    OptDepends,
    Packager,
    PgpSig,
    Provides,
    Replaces,
    SchemaVersionV2,
    Sha256Sum,
    Url,
    Version,
):
    """A model describing all identifiers in a 'desc' file (version 2).

    Attributes
    ----------
    arch: str
        The attribute can be used to describe the (required) data below an %ARCH% identifier in a 'desc' file, which
        identifies a package's architecture
    backup: list[str] | None
        The attribute can be used to describe the (optional) data below a %BACKUP% identifier in a 'desc' file, which
        identifies which file(s) of a package pacman will create backups for
    base: str
        The attribute can be used to describe the (required) data below a %BASE% identifier in a 'desc' file, which
        identifies a package's pkgbase
    builddate: int
        The attribute can be used to describe the (required) data below a %BUILDDATE% identifier in a 'desc' file,
        which identifies a package's build date (represented in seconds since the epoch)
    checkdepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %CHECKDEPENDS% identifier in a 'desc' file,
        which identifies a package's checkdepends
    conflicts: list[str] | None
        The attribute can be used to describe the (optional) data below a %CONFLICTS% identifier in a 'desc' file, which
        identifies what other package(s) a package conflicts with
    csize: int
        The attribute can be used to describe the (required) data below a %CSIZE% identifier in a 'desc' file, which
        identifies a package's size
    depends: list[str] | None
        The attribute can be used to describe the (optional) data below a %DEPENDS% identifier in a 'desc' file, which
        identifies what other package(s) a package depends on
    desc: str
        The attribute can be used to describe the (required) data below a %DESC% identifier in a 'desc' file, which
        identifies a package's description
    filename: str
        The attribute can be used to describe the (required) data below a %FILENAME% identifier in a 'desc' file, which
        identifies a package's file name
    groups: list[str] | None
        The attribute can be used to describe the (optional) data below a %GROUPS% identifier in a 'desc' file, which
        identifies a package's groups
    isize: int
        The attribute can be used to describe the (required) data below an %ISIZE% identifier in a 'desc' file, which
        identifies a package's installed size
    license: list[str]
        The attribute can be used to describe the (required) data below a %LICENSE% identifier in a 'desc' file, which
        identifies a package's license(s)
    makedepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %MAKEDEPENDS% identifier in a 'desc' file,
        which identifies a package's makedepends
    name: str
        The attribute can be used to describe the (required) data below a %NAME% identifier in a 'desc' file, which
        identifies a package's name
    optdepends: list[str] | None
        The attribute can be used to describe the (optional) data below a %OPTDEPENDS% identifier in a 'desc' file,
        which identifies what other package(s) a package optionally depends on
    packager: str
        The attribute can be used to describe the (required) data below a %PACKAGER% identifier in a 'desc' file, which
        identifies a package's packager
    pgpsig: str | None
        The attribute can be used to describe the (optional) data below a %PGPSIG% identifier in a 'desc' file, which
        identifies a package's PGP signature
    provides: list[str] | None
        The attribute can be used to describe the (optional) data below a %PROVIDES% identifier in a 'desc' file, which
        identifies what other package(s) a package provides
    replaces: list[str] | None
        The attribute can be used to describe the (optional) data below a %REPLACES% identifier in a 'desc' file, which
        identifies what other package(s) a package replaces
    schema_version: PositiveInt
        A positive integer - 1 - identifying the schema version of the object
    sha256sum: str
        The attribute can be used to describe the (required) data below an %SHA256SUM% identifier in a 'desc' file,
        which identifies a package's sha256 checksum
    url: str
        The attribute can be used to describe the (required) data below a %URL% identifier in a 'desc' file, which
        identifies a package's URL
    version: str
        The attribute can be used to describe the (required) data below a %VERSION% identifier in a 'desc' file, which
        identifies a package's version (this is the accumulation of epoch, pkgver and pkgrel)
    """


def export_schemas(output: Path | str) -> None:
    """Export the JSON schema of selected pydantic models to an output directory.

    Parameters
    ----------
    output: Path
        A path to which to output the JSON schema files

    Raises
    ------
    RuntimeError
        If output is not an existing directory
    """
    classes = [FilesV1, PackageDescV1, PackageDescV2]

    if isinstance(output, str):
        output = Path(output)

    if not output.exists():
        raise RuntimeError(f"The output directory {output} must exist!")

    for class_ in classes:
        with open(output / f"{class_.__name__}.json", "wb") as f:
            f.write(
                dumps(
                    class_.model_json_schema(),  # type: ignore[attr-defined]
                    option=OPT_INDENT_2 | OPT_APPEND_NEWLINE | OPT_SORT_KEYS,
                )
            )



"""An implementation of libalpm functionality for version comparison."""

from itertools import takewhile

from .util import cmp

PYALPM_VERCMP = False
try:
    from pyalpm import vercmp as pyalpm_vercmp  # pragma: no-cover-nonlinux

    PYALPM_VERCMP = True  # pragma: no-cover-nonlinux
except ImportError:  # pragma: nocover
    pass


def vercmp(a: str, b: str) -> int:  # noqa: C901
    """Compare alpha and numeric segments of two versions.

    The comparison algorithm is based on libalpm pacman's vercmp behavior.

    Calls pyalpm's interface in case it's available.

    Returns
    -------
    int
        -1 if a is newer than b
         0 if a and b are the same version
         1 if b is newer than a
    """
    if PYALPM_VERCMP:  # pragma: no-cover-nonlinux
        return int(pyalpm_vercmp(a, b))  # pragma: no-cover-nonlinux

    # easy comparison to see if versions are identical
    if a == b:
        return 0

    one: str
    two: str

    one_ptr = ptr1 = 0
    two_ptr = ptr2 = 0

    # loop through each version segment of a and b and compare them
    while one_ptr < len(a) and two_ptr < len(b):
        while one_ptr < len(a) and not a[one_ptr].isalnum():
            one_ptr += 1
        while two_ptr < len(b) and not b[two_ptr].isalnum():
            two_ptr += 1

        # If we ran to the end of either, we are finished with the loop
        if one_ptr >= len(a) or two_ptr >= len(b):
            break

        # If the separator lengths were different, we are finished
        if (one_ptr - ptr1) != (two_ptr - ptr2):
            return -1 if (one_ptr - ptr1) < (two_ptr - ptr2) else 1

        # adjust left side pointer to current segment start
        ptr1 = one_ptr
        ptr2 = two_ptr
        one = a[one_ptr:]
        two = b[two_ptr:]

        """
        grab first completely alpha or completely numeric segment
        leave one and two pointing to the start of the alpha or numeric
        segment and walk ptr1 and ptr2 to end of segment
        """
        if one and one[0].isdecimal():
            ptr1 += len(list(takewhile(str.isdecimal, one)))
            ptr2 += len(list(takewhile(str.isdecimal, two)))
            isnum = True
        else:
            ptr1 += len(list(takewhile(str.isalpha, one)))
            ptr2 += len(list(takewhile(str.isalpha, two)))
            isnum = False

        # adjust current segment end with the updated right side pointer
        one = a[one_ptr:ptr1]
        two = b[two_ptr:ptr2]

        """
        take care of the case where the two version segments are
        different types: one numeric, the other alpha (i.e. empty)
        numeric segments are always newer than alpha segments
        """
        if not two:
            return 1 if isnum else -1

        if isnum:
            # throw away any leading zeros - it's a number, right?
            one = one.lstrip("0")
            two = two.lstrip("0")

            # whichever number has more digits wins (discard leading zeros)
            if len(one) > len(two):
                return 1
            if len(two) > len(one):
                return -1

        """
        strcmp will return which one is greater - even if the two
        segments are alpha or if they are numeric.  don't return
        if they are equal because there might be more segments to
        compare
        """
        rc = cmp(one, two)
        if rc:
            return rc

        # advance left side pointer to current right side pointer
        one_ptr = ptr1
        two_ptr = ptr2

    # set leftover using the left side pointer once the segment loop finished
    one = a[one_ptr:]
    two = b[two_ptr:]

    """
    this catches the case where all numeric and alpha segments have
    compared identically but the segment separating characters were
    different
    """
    if not one and not two:
        return 0

    """
    the final showdown. we never want a remaining alpha string to
    beat an empty string. the logic is a bit weird, but:
    - if one is empty and two is not an alpha, two is newer.
    - if one is an alpha, two is newer.
    - otherwise one is newer.
    """
    if (not one and not two[0].isalpha()) or (one and one[0].isalpha()):
        return -1
    return 1


def pkg_vercmp(a: str, b: str) -> int:
    """Compare individual components of versions by splitting based on alpm's version scheme into epoch, pkgver, pkgrel.

    The comparison algorithm is based on libalpm pacman's vercmp behavior.

    Calls pyalpm's interface in case it's available.

    Returns
    -------
    int
        -1 if a is newer than b
         0 if a and b are the same version
         1 if b is newer than a
    """
    if PYALPM_VERCMP:  # pragma: no-cover-nonlinux
        return int(pyalpm_vercmp(a, b))  # pragma: no-cover-nonlinux

    epoch1 = a.split(":")[0] if ":" in a else ""
    epoch2 = b.split(":")[0] if ":" in b else ""

    pkgver_pkgrel1 = a.split(":")[1] if ":" in a else a
    pkgver_pkgrel2 = b.split(":")[1] if ":" in b else b

    pkgver1 = pkgver_pkgrel1.split("-")[0]
    pkgver2 = pkgver_pkgrel2.split("-")[0]

    pkgrel1 = pkgver_pkgrel1.split("-")[1]
    pkgrel2 = pkgver_pkgrel2.split("-")[1]

    # return epoch comparison if unequal
    ret = cmp(epoch1, epoch2)
    if ret:
        return ret

    # return pkgver comparison if unequal
    ret = vercmp(pkgver1, pkgver2)
    if ret:
        return ret

    # return pkgrel comparison
    return vercmp(pkgrel1, pkgrel2)
